"""Collection of callback utils to run common OTX algorithms."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import time

from otx.api.usecases.reporting.time_monitor_callback import TimeMonitorCallback


class TrainingProgressCallback(TimeMonitorCallback):
    """TrainingProgressCallback class for time monitoring."""

    def __init__(self, update_progress_callback, **kwargs):
        super().__init__(update_progress_callback=update_progress_callback, **kwargs)

    def on_train_batch_end(self, batch, logs=None):
        """Callback function on training batch ended."""
        super().on_train_batch_end(batch, logs)
        self.update_progress_callback(self.get_progress())

    def on_epoch_end(self, epoch, logs=None):
        """Callback function on epoch ended."""
        self.past_epoch_duration.append(time.time() - self.start_epoch_time)
        progress = ((epoch + 1) / self.total_epochs) * 100
        self._calculate_average_epoch()
        score = None
        if hasattr(self.update_progress_callback, "metric") and isinstance(logs, dict):
            score = logs.get(self.update_progress_callback.metric, None)
        self.update_progress_callback(progress, score=score)


class InferenceProgressCallback(TimeMonitorCallback):
    """InferenceProgressCallback class for time monitoring."""

    def __init__(self, num_test_steps, update_progress_callback, **kwargs):
        super().__init__(
            num_epoch=0,
            num_train_steps=0,
            num_val_steps=0,
            num_test_steps=num_test_steps,
            update_progress_callback=update_progress_callback,
            **kwargs,
        )

    def on_test_batch_end(self, batch=None, logs=None):
        """Callback function on testing batch ended."""
        super().on_test_batch_end(batch, logs)
        self.update_progress_callback(int(self.get_progress()))


class OptimizationProgressCallback(TrainingProgressCallback):
    """Progress callback used for optimization using NNCF.

    There are three stages to the progress bar:
       - 5 % model is loaded
       - 10 % compressed model is initialized
       - 10-100 % compressed model is being fine-tuned
    """

    def __init__(
        self,
        update_progress_callback,
        loading_stage_progress_percentage: int = 5,
        initialization_stage_progress_percentage: int = 5,
        **kwargs,
    ):
        super().__init__(update_progress_callback=update_progress_callback, **kwargs)
        if loading_stage_progress_percentage + initialization_stage_progress_percentage >= 100:
            raise RuntimeError("Total optimization progress percentage is more than 100%")

        self.loading_stage_progress_percentage = loading_stage_progress_percentage
        self.initialization_stage_progress_percentage = initialization_stage_progress_percentage

        # set loading_stage_progress_percentage from the start as the model is already loaded at this point
        if self.update_progress_callback:
            self.update_progress_callback(loading_stage_progress_percentage)

    def on_train_begin(self, logs=None):
        """Callback function when training beginning."""
        super().on_train_begin(logs)
        # Callback initialization takes place here after OTXProgressHook.before_run() is called
        train_percentage = 100 - self.loading_stage_progress_percentage - self.initialization_stage_progress_percentage
        loading_stage_steps = self.total_steps * self.loading_stage_progress_percentage / train_percentage
        initialization_stage_steps = self.total_steps * self.initialization_stage_progress_percentage / train_percentage
        self.total_steps += loading_stage_steps + initialization_stage_steps

        self.current_step = loading_stage_steps + initialization_stage_steps
        self.update_progress_callback(self.get_progress())

    def on_train_end(self, logs=None):
        """Callback function on training ended."""
        super().on_train_end(logs)
        self.update_progress_callback(self.get_progress(), score=logs)

    def on_initialization_end(self):
        """on_initialization_end callback for optimization using NNCF."""
        self.update_progress_callback(
            self.loading_stage_progress_percentage + self.initialization_stage_progress_percentage
        )



"""Averaging module contains averaging method enumeration."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from enum import Enum, auto


class MetricAverageMethod(Enum):
    """This defines the metrics averaging method."""

    MICRO = auto()
    MACRO = auto()



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest

from otx.algorithms.classification.tasks import ClassificationInferenceTask
from otx.algorithms.common.tasks import BaseTask
from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label import Domain
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.metrics import Performance, ScoreMetric
from otx.api.entities.model import ModelConfiguration, ModelEntity
from otx.api.entities.resultset import ResultSetEntity
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.classification.test_helper import (
    DEFAULT_CLS_TEMPLATE,
    init_environment,
    setup_configurable_parameters,
)


@pytest.fixture
def otx_classification_model():
    model_configuration = ModelConfiguration(
        configurable_parameters=ConfigurableParameters(header="header", description="description"),
        label_schema=LabelSchemaEntity(),
    )
    return ModelEntity(train_dataset=DatasetEntity(), configuration=model_configuration)


class TestOTXClassificationInferenceTask:
    @pytest.fixture(autouse=True)
    def setup(self, otx_classification_model, tmp_dir_path) -> None:
        self.dataset_len = 5
        self.hyper_parameters, self.model_template = setup_configurable_parameters(DEFAULT_CLS_TEMPLATE)
        task_environment, self.dataset = init_environment(
            self.hyper_parameters, self.model_template, False, False, self.dataset_len
        )

        self.output_path = str(tmp_dir_path)
        self.cls_inference_task = ClassificationInferenceTask(task_environment, output_path=self.output_path)
        self.model = otx_classification_model

    @pytest.mark.parametrize("multilabel, hierarchical", [(False, False), (True, False), (False, True)])
    @e2e_pytest_unit
    def test_infer(self, mocker, multilabel, hierarchical):
        task_environment, dataset = init_environment(
            self.hyper_parameters, self.model_template, multilabel, hierarchical, self.dataset_len
        )
        custom_cls_inference_task = ClassificationInferenceTask(task_environment, output_path=self.output_path)

        items_num = len(dataset)
        fake_output = {
            "outputs": {
                "eval_predictions": np.zeros((items_num, 5)),
                "feature_vectors": np.zeros((items_num, 5)),
                "saliency_maps": np.zeros((items_num, 5, 5)).astype(np.uint8),
            }
        }

        mock_run_task = mocker.patch.object(BaseTask, "_run_task", return_value=fake_output)
        inf_params = InferenceParameters()
        inf_params.is_evaluation = True
        updated_dataset = custom_cls_inference_task.infer(dataset.with_empty_annotations(), inf_params)

        mock_run_task.assert_called_once()
        for updated in updated_dataset:
            if not multilabel and not hierarchical:
                assert len(updated.annotation_scene.get_labels()) == 1
            for lbl in updated.annotation_scene.get_labels():
                assert lbl.domain == Domain.CLASSIFICATION

    @e2e_pytest_unit
    def test_explain(self, mocker):
        items_num = len(self.dataset)
        fake_output = {
            "outputs": {
                "eval_predictions": np.zeros((items_num, 5)),
                "feature_vectors": np.zeros((items_num, 5)),
                "saliency_maps": np.zeros((items_num, 5, 5)).astype(np.uint8),
            }
        }

        mock_run_task = mocker.patch.object(BaseTask, "_run_task", return_value=fake_output)
        updated_dataset = self.cls_inference_task.explain(self.dataset.with_empty_annotations())
        mock_run_task.assert_called_once()

        for item in updated_dataset:
            data_list = item.get_metadata()
            for data in data_list:
                assert data.data.type == "saliency_map"
                assert data.data.numpy is not None

    @e2e_pytest_unit
    def test_evaluate(self, mocker):
        result_set = ResultSetEntity(
            model=self.model,
            ground_truth_dataset=DatasetEntity(),
            prediction_dataset=DatasetEntity(),
        )
        fake_metrics = mocker.patch("otx.api.usecases.evaluation.accuracy.Accuracy", autospec=True)
        fake_metrics.get_performance.return_value = Performance(
            score=ScoreMetric(name="fake", value=0.1), dashboard_metrics="accuracy"
        )
        mocker.patch.object(MetricsHelper, "compute_accuracy", return_value=fake_metrics)
        self.cls_inference_task.evaluate(result_set)

        assert result_set.performance.score.value == 0.1

    @e2e_pytest_unit
    def test_export_empty_model(self, mocker):
        fake_output = {"outputs": {"bin": None, "xml": None}}
        mock_run_task = mocker.patch.object(BaseTask, "_run_task", return_value=fake_output)

        with pytest.raises(RuntimeError):
            self.cls_inference_task.export(ExportType.OPENVINO, self.model)
            mock_run_task.assert_called_once()

    @e2e_pytest_unit
    def test_export_with_model_files(self, mocker):
        with open(f"{self.output_path}/model.xml", "wb") as f:
            f.write(b"foo")
        with open(f"{self.output_path}/model.bin", "wb") as f:
            f.write(b"bar")

        fake_output = {"outputs": {"bin": f"{self.output_path}/model.xml", "xml": f"{self.output_path}/model.bin"}}
        mock_run_task = mocker.patch.object(BaseTask, "_run_task", return_value=fake_output)
        self.cls_inference_task.export(ExportType.OPENVINO, self.model)

        mock_run_task.assert_called_once()
        assert self.model.get_data("openvino.bin")
        assert self.model.get_data("openvino.xml")

    @e2e_pytest_unit
    def test_unload(self, mocker):
        mock_cleanup = mocker.patch.object(BaseTask, "cleanup")
        self.cls_inference_task.unload()

        mock_cleanup.assert_called_once()



"""Data Pipeline of X3D model for Action Cls Task."""

# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name
# dataset settings
seed = 2
dataset_type = "RawframeDataset"

img_norm_cfg = dict(mean=[0.0, 0.0, 0.0], std=[255.0, 255.0, 255.0], to_bgr=False)

clip_len = 8
frame_interval = 4
train_pipeline = [
    dict(type="SampleFrames", clip_len=clip_len, frame_interval=frame_interval, num_clips=1),
    dict(type="RawFrameDecode"),
    dict(type="Resize", scale=(-1, 256)),
    dict(type="RandomResizedCrop"),
    dict(type="Resize", scale=(224, 224), keep_ratio=False),
    dict(type="Flip", flip_ratio=0.5),
    dict(type="Normalize", **img_norm_cfg),
    dict(type="FormatShape", input_format="NCTHW"),
    dict(type="Collect", keys=["imgs", "label"], meta_keys=[]),
    dict(type="ToTensor", keys=["imgs", "label"]),
]

val_pipeline = [
    dict(type="SampleFrames", clip_len=clip_len, frame_interval=frame_interval, num_clips=1, test_mode=True),
    dict(type="RawFrameDecode"),
    dict(type="Resize", scale=(-1, 256)),
    dict(type="CenterCrop", crop_size=224),
    dict(type="Normalize", **img_norm_cfg),
    dict(type="FormatShape", input_format="NCTHW"),
    dict(type="Collect", keys=["imgs", "label"], meta_keys=[]),
    dict(type="ToTensor", keys=["imgs"]),
]
# TODO Delete label in meta key in test pipeline
test_pipeline = [
    dict(type="SampleFrames", clip_len=clip_len, frame_interval=frame_interval, num_clips=1, test_mode=True),
    dict(type="RawFrameDecode"),
    dict(type="Resize", scale=(-1, 256)),
    dict(type="CenterCrop", crop_size=224),
    dict(type="Normalize", **img_norm_cfg),
    dict(type="FormatShape", input_format="NCTHW"),
    dict(type="Collect", keys=["imgs"], meta_keys=[]),
    dict(type="ToTensor", keys=["imgs"]),
]

data = dict(
    videos_per_gpu=10,
    workers_per_gpu=0,
    val_dataloader=dict(videos_per_gpu=1),
    test_dataloader=dict(videos_per_gpu=1),
    train=dict(
        type=dataset_type,
        pipeline=train_pipeline,
    ),
    val=dict(
        type=dataset_type,
        pipeline=val_pipeline,
    ),
    test=dict(
        type=dataset_type,
        pipeline=test_pipeline,
    ),
)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import functools

import torch
from mmdet.models.builder import DETECTORS
from mmdet.models.detectors.mask_rcnn import MaskRCNN

from otx.mpa.deploy.utils import is_mmdeploy_enabled
from otx.mpa.modules.utils.task_adapt import map_class_names
from otx.mpa.utils.logger import get_logger

from .l2sp_detector_mixin import L2SPDetectorMixin
from .sam_detector_mixin import SAMDetectorMixin

logger = get_logger()


@DETECTORS.register_module()
class CustomMaskRCNN(SAMDetectorMixin, L2SPDetectorMixin, MaskRCNN):
    def __init__(self, *args, task_adapt=None, **kwargs):
        super().__init__(*args, **kwargs)

        # Hook for class-sensitive weight loading
        if task_adapt:
            self._register_load_state_dict_pre_hook(
                functools.partial(
                    self.load_state_dict_pre_hook,
                    self,  # model
                    task_adapt["dst_classes"],  # model_classes
                    task_adapt["src_classes"],  # chkpt_classes
                )
            )

    @staticmethod
    def load_state_dict_pre_hook(model, model_classes, chkpt_classes, chkpt_dict, prefix, *args, **kwargs):
        """Modify input state_dict according to class name matching before weight loading"""
        logger.info(f"----------------- CustomMaskRCNN.load_state_dict_pre_hook() called w/ prefix: {prefix}")

        # Dst to src mapping index
        model_dict = model.state_dict()
        model_classes = list(model_classes)
        chkpt_classes = list(chkpt_classes)
        model2chkpt = map_class_names(model_classes, chkpt_classes)
        logger.info(f"{chkpt_classes} -> {model_classes} ({model2chkpt})")

        # List of class-relevant params & their row-stride
        param_strides = {
            "roi_head.bbox_head.fc_cls.weight": 1,
            "roi_head.bbox_head.fc_cls.bias": 1,
            "roi_head.bbox_head.fc_reg.weight": 4,  # 4 rows (bbox) for each class
            "roi_head.bbox_head.fc_reg.bias": 4,
        }

        for model_name, stride in param_strides.items():
            chkpt_name = prefix + model_name
            if model_name not in model_dict or chkpt_name not in chkpt_dict:
                logger.info(f"Skipping weight copy: {chkpt_name}")
                continue

            # Mix weights
            model_param = model_dict[model_name].clone()
            chkpt_param = chkpt_dict[chkpt_name]
            for m, c in enumerate(model2chkpt):
                if c >= 0:
                    # Copying only matched weight rows
                    model_param[(m) * stride : (m + 1) * stride].copy_(chkpt_param[(c) * stride : (c + 1) * stride])
            if model_param.shape[0] > len(model_classes * stride):  # BG class
                c = len(chkpt_classes)
                m = len(model_classes)
                model_param[(m) * stride : (m + 1) * stride].copy_(chkpt_param[(c) * stride : (c + 1) * stride])

            # Replace checkpoint weight by mixed weights
            chkpt_dict[chkpt_name] = model_param


if is_mmdeploy_enabled():
    from mmdeploy.core import FUNCTION_REWRITER, mark
    from mmdeploy.utils import is_dynamic_shape

    from otx.mpa.modules.hooks.recording_forward_hooks import (
        ActivationMapHook,
        FeatureVectorHook,
    )

    @FUNCTION_REWRITER.register_rewriter(
        "otx.mpa.modules.models.detectors.custom_maskrcnn_detector.CustomMaskRCNN.simple_test"
    )
    def custom_mask_rcnn__simple_test(ctx, self, img, img_metas, proposals=None, **kwargs):
        assert self.with_bbox, "Bbox head must be implemented."
        x = self.extract_feat(img)
        if proposals is None:
            proposals, _ = self.rpn_head.simple_test_rpn(x, img_metas)
        out = self.roi_head.simple_test(x, proposals, img_metas, rescale=False)

        if ctx.cfg["dump_features"]:
            feature_vector = FeatureVectorHook.func(x)
            saliency_map = ActivationMapHook.func(x[-1])
            return (*out, feature_vector, saliency_map)

        return out

    @mark("custom_maskrcnn_forward", inputs=["input"], outputs=["dets", "labels", "masks", "feats", "saliencies"])
    def __forward_impl(ctx, self, img, img_metas, **kwargs):
        assert isinstance(img, torch.Tensor)

        deploy_cfg = ctx.cfg
        is_dynamic_flag = is_dynamic_shape(deploy_cfg)
        # get origin input shape as tensor to support onnx dynamic shape
        img_shape = torch._shape_as_tensor(img)[2:]
        if not is_dynamic_flag:
            img_shape = [int(val) for val in img_shape]
        img_metas[0]["img_shape"] = img_shape
        return self.simple_test(img, img_metas, **kwargs)

    @FUNCTION_REWRITER.register_rewriter(
        "otx.mpa.modules.models.detectors.custom_maskrcnn_detector.CustomMaskRCNN.forward"
    )
    def custom_maskrcnn__forward(ctx, self, img, img_metas=None, return_loss=False, **kwargs):
        if img_metas is None:
            img_metas = [{}]
        else:
            assert len(img_metas) == 1, "do not support aug_test"
            img_metas = img_metas[0]

        if isinstance(img, list):
            img = img[0]

        return __forward_impl(ctx, self, img, img_metas=img_metas, **kwargs)



from mmdet.models.builder import HEADS
from mmdet.models.dense_heads.yolox_head import YOLOXHead


@HEADS.register_module()
class CustomYOLOXHead(YOLOXHead):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# import torch
import numpy as np
from mmdet.datasets import DATASETS, PIPELINES, build_dataset

from otx.mpa.modules.utils.task_adapt import map_cat_and_cls_as_order, map_class_names


@DATASETS.register_module()
class TaskAdaptEvalDataset:
    """Dataset wrapper for task-adaptive evaluation.

    Attributes:
        model_classes:
        **kwargs:

    Notes:
        TODO[Eugene/Jaeguk]: check if this class is still valid in OTX
    """

    def __init__(self, model_classes, **kwargs):
        dataset_cfg = kwargs.copy()
        org_type = dataset_cfg.pop("org_type")
        dataset_cfg["type"] = org_type
        self.dataset = build_dataset(dataset_cfg)
        self.model_classes = model_classes
        self.CLASSES = self.dataset.CLASSES
        self.data2model = map_class_names(self.CLASSES, self.model_classes)
        if org_type == "CocoDataset":
            self.dataset.cat2label, self.dataset.cat_ids = map_cat_and_cls_as_order(
                self.CLASSES, self.dataset.coco.cats
            )

    def __getitem__(self, idx):
        return self.dataset[idx]

    def __len__(self):
        return len(self.dataset)

    def evaluate(self, results, **kwargs):
        # Filter & reorder detection results
        adapt_results = []
        for result in results:  # for each image
            adapt_result = []
            for model_class_index in self.data2model:  # for each class
                # Gather per-class results according to index mapping
                if model_class_index >= 0:
                    adapt_result.append(result[model_class_index])
                else:
                    adapt_result.append(np.empty([0, 5]))
            adapt_results.append(adapt_result)

        # Call evaluation w/ org arguments
        return self.dataset.evaluate(adapt_results, **kwargs)


@PIPELINES.register_module()
class AdaptClassLabels(object):
    """Data processor for task-adative annotation loading."""

    def __init__(self, src_classes, dst_classes):
        self.src2dst = map_class_names(src_classes, dst_classes)
        print("AdaptClassLabels")
        print("src_classes", src_classes)
        print("dst_classes", dst_classes)
        print("src2dst", self.src2dst)

    def __call__(self, data):
        src_labels = data["gt_labels"]
        dst_labels = []
        for src_label in src_labels:
            dst_labels.append(self.src2dst[src_label])
        data["gt_labels"] = np.array(dst_labels)
        return data



_base_ = "../model.py"

model = dict(
    type="ImageClassifier",
    task="classification",
    pretrained=None,
    backbone=dict(),
    neck=dict(type="GlobalAveragePooling"),
    head=dict(in_channels=-1, loss=dict(type="CrossEntropyLoss", loss_weight=1.0), topk=(1, 5)),
)

checkpoint_config = dict(type="CheckpointHookWithValResults")



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv import ConfigDict

from otx.algorithms.segmentation.adapters.mmseg.utils.builder import build_segmentor
from otx.mpa.stage import Stage
from otx.mpa.utils.config_utils import recursively_update_cfg
from otx.mpa.utils.logger import get_logger

logger = get_logger()


class SegStage(Stage):
    MODEL_BUILDER = build_segmentor

    def configure(self, model_cfg, model_ckpt, data_cfg, training=True, **kwargs):
        """Create MMCV-consumable config from given inputs"""
        logger.info(f"configure!: training={training}")

        cfg = self.cfg
        self.configure_model(cfg, model_cfg, training, **kwargs)
        self.configure_ckpt(cfg, model_ckpt, kwargs.get("pretrained", None))
        self.configure_data(cfg, training, data_cfg)
        self.configure_task(cfg, training, **kwargs)
        self.configure_hook(cfg)

        return cfg

    def configure_model(self, cfg, model_cfg, training, **kwargs):

        if model_cfg:
            if hasattr(model_cfg, "model"):
                cfg.merge_from_dict(model_cfg._cfg_dict)
            else:
                raise ValueError(
                    "Unexpected config was passed through 'model_cfg'. "
                    "it should have 'model' attribute in the config"
                )
            cfg.model_task = cfg.model.pop("task", "segmentation")
            if cfg.model_task != "segmentation":
                raise ValueError(f"Given model_cfg ({model_cfg.filename}) is not supported by segmentation recipe")

        # OV-plugin
        ir_model_path = kwargs.get("ir_model_path")
        if ir_model_path:

            def is_mmov_model(k, v):
                if k == "type" and v.startswith("MMOV"):
                    return True
                return False

            ir_weight_path = kwargs.get("ir_weight_path", None)
            ir_weight_init = kwargs.get("ir_weight_init", False)
            recursively_update_cfg(
                cfg,
                is_mmov_model,
                {"model_path": ir_model_path, "weight_path": ir_weight_path, "init_weight": ir_weight_init},
            )

    def configure_data(self, cfg, training, data_cfg, **kwargs):  # noqa: C901
        # Data
        if data_cfg:
            cfg.merge_from_dict(data_cfg)

        # Dataset
        super().configure_data(cfg, training, **kwargs)
        src_data_cfg = Stage.get_data_cfg(cfg, "train")
        for mode in ["train", "val", "test"]:
            if src_data_cfg.type == "MPASegDataset" and cfg.data.get(mode, False):
                if cfg.data[mode]["type"] != "MPASegDataset":
                    # Wrap original dataset config
                    org_type = cfg.data[mode]["type"]
                    cfg.data[mode]["type"] = "MPASegDataset"
                    cfg.data[mode]["org_type"] = org_type

    def configure_task(self, cfg, training, **kwargs):
        """Adjust settings for task adaptation"""
        if cfg.get("task_adapt", None):
            logger.info(f"task config!!!!: training={training}")
            task_adapt_op = cfg["task_adapt"].get("op", "REPLACE")

            # Task classes
            self.configure_classes(cfg, task_adapt_op)
            # Ignored mode
            self.configure_ignore(cfg)

    def configure_classes(self, cfg, task_adapt_op):
        # Task classes
        org_model_classes = self.get_model_classes(cfg)
        data_classes = self.get_data_classes(cfg)
        if "background" not in org_model_classes:
            org_model_classes = ["background"] + org_model_classes
        if "background" not in data_classes:
            data_classes = ["background"] + data_classes

        # Model classes
        if task_adapt_op == "REPLACE":
            if len(data_classes) == 1:  # 'background'
                model_classes = org_model_classes.copy()
            else:
                model_classes = data_classes.copy()
        elif task_adapt_op == "MERGE":
            model_classes = org_model_classes + [cls for cls in data_classes if cls not in org_model_classes]
        else:
            raise KeyError(f"{task_adapt_op} is not supported for task_adapt options!")

        cfg.task_adapt.final = model_classes
        cfg.model.task_adapt = ConfigDict(
            src_classes=org_model_classes,
            dst_classes=model_classes,
        )

        # Model architecture
        if "decode_head" in cfg.model:
            decode_head = cfg.model.decode_head
            if isinstance(decode_head, dict):
                decode_head.num_classes = len(model_classes)
            elif isinstance(decode_head, list):
                for head in decode_head:
                    head.num_classes = len(model_classes)

            # For SupConDetCon
            if "SupConDetCon" in cfg.model.type:
                cfg.model.num_classes = len(model_classes)

        # Task classes
        self.org_model_classes = org_model_classes
        self.model_classes = model_classes

    def configure_ignore(self, cfg):
        # Change to incremental loss (ignore mode)
        if cfg.get("ignore", False):
            cfg_loss_decode = ConfigDict(
                type="CrossEntropyLossWithIgnore",
                use_sigmoid=False,
                loss_weight=1.0,
            )

            if "decode_head" in cfg.model:
                decode_head = cfg.model.decode_head
                if decode_head.type == "FCNHead":
                    decode_head.type = "CustomFCNHead"
                    decode_head.loss_decode = cfg_loss_decode



"""API Tests for Action Detection training"""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import glob
import os.path as osp
import time
from concurrent.futures import ThreadPoolExecutor
from typing import Optional

import numpy as np
import pytest

from otx.algorithms.action.tasks import ActionInferenceTask, ActionTrainTask
from otx.algorithms.common.tasks.training_base import BaseTask
from otx.api.configuration.helper import create
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.metrics import Performance
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import parse_model_template
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.train_parameters import TrainParameters
from otx.core.data.adapter import get_dataset_adapter
from tests.test_suite.e2e_test_system import e2e_pytest_api

DEFAULT_ACTION_TEMPLATE_DIR = osp.join("otx/algorithms/action/configs", "detection", "x3d_fast_rcnn")


def task_eval(task: BaseTask, model: ModelEntity, dataset: DatasetEntity) -> Performance:
    start_time = time.time()
    result_dataset = task.infer(dataset.with_empty_annotations())
    end_time = time.time()
    print(f"{len(dataset)} analysed in {end_time - start_time} seconds")
    result_set = ResultSetEntity(model=model, ground_truth_dataset=dataset, prediction_dataset=result_dataset)
    task.evaluate(result_set)
    assert result_set.performance is not None
    return result_set.performance


class TestActionTaskAPI:
    """
    Collection of tests for OTX API and OTX Model Templates
    """

    train_data_roots = "tests/assets/cvat_dataset/action_detection/train"
    val_data_roots = "tests/assets/cvat_dataset/action_detection/train"

    @e2e_pytest_api
    def test_reading_action_model_template(self):
        model_templates = ["x3d_fast_rcnn"]
        for model_template in model_templates:
            parse_model_template(
                osp.join("otx/algorithms/action/configs", "detection", model_template, "template.yaml")
            )

    def init_environment(self, params, model_template):
        dataset_adapter = get_dataset_adapter(
            model_template.task_type,
            train_data_roots=self.train_data_roots,
            val_data_roots=self.val_data_roots,
        )
        dataset = dataset_adapter.get_otx_dataset()
        label_schema = dataset_adapter.get_label_schema()
        environment = TaskEnvironment(
            model=None,
            hyper_parameters=params,
            label_schema=label_schema,
            model_template=model_template,
        )
        return environment, dataset

    @staticmethod
    def setup_configurable_parameters(template_dir, num_iters=10):
        glb = glob.glob(f"{template_dir}/template*.yaml")
        template_path = glb[0] if glb else None
        if not template_path:
            raise RuntimeError(f"Template YAML not found: {template_dir}")

        model_template = parse_model_template(template_path)
        hyper_parameters = create(model_template.hyper_parameters.data)
        hyper_parameters.learning_parameters.num_iters = num_iters
        hyper_parameters.postprocessing.result_based_confidence_threshold = False
        hyper_parameters.postprocessing.confidence_threshold = 0.1
        return hyper_parameters, model_template

    @e2e_pytest_api
    @pytest.mark.skip(reason="mmaction does not support EpochRunnerWithCancel")
    def test_cancel_training_action(self):
        """
        Tests starting and cancelling training.

        Flow of the test:
        - Creates a randomly annotated project with a small dataset containing 3 classes:
            ['rectangle', 'triangle', 'circle'].
        - Start training and give cancel training signal after 10 seconds. Assert that training
            stops within 35 seconds after that
        - Start training and give cancel signal immediately. Assert that training stops within 25 seconds.

        This test should be finished in under one minute on a workstation.
        """
        hyper_parameters, model_template = self.setup_configurable_parameters(
            DEFAULT_ACTION_TEMPLATE_DIR, num_iters=500
        )
        action_environment, dataset = self.init_environment(hyper_parameters, model_template)

        action_task = ActionTrainTask(task_environment=action_environment)

        executor = ThreadPoolExecutor(max_workers=1, thread_name_prefix="train_thread")

        output_model = ModelEntity(
            dataset,
            action_environment.get_model_configuration(),
        )

        training_progress_curve = []

        def progress_callback(progress: float, score: Optional[float] = None):
            training_progress_curve.append(progress)

        train_parameters = TrainParameters()
        train_parameters.update_progress = progress_callback

        # Test stopping after some time
        start_time = time.time()
        train_future = executor.submit(action_task.train, dataset, output_model, train_parameters)
        # give train_thread some time to initialize the model
        while not action_task._is_training:
            time.sleep(10)
        action_task.cancel_training()

        # stopping process has to happen in less than 35 seconds
        train_future.result()
        assert training_progress_curve[-1] == 100
        assert time.time() - start_time < 100, "Expected to stop within 100 seconds."

        # Test stopping immediately
        start_time = time.time()
        train_future = executor.submit(action_task.train, dataset, output_model)
        action_task.cancel_training()

        train_future.result()
        assert time.time() - start_time < 25  # stopping process has to happen in less than 25 seconds

    @e2e_pytest_api
    def test_training_progress_tracking(self):
        hyper_parameters, model_template = self.setup_configurable_parameters(DEFAULT_ACTION_TEMPLATE_DIR, num_iters=5)
        action_environment, dataset = self.init_environment(hyper_parameters, model_template)
        task = ActionTrainTask(task_environment=action_environment)
        print("Task initialized, model training starts.")

        training_progress_curve = []

        def progress_callback(progress: float, score: Optional[float] = None):
            training_progress_curve.append(progress)

        train_parameters = TrainParameters()
        train_parameters.update_progress = progress_callback
        output_model = ModelEntity(
            dataset,
            action_environment.get_model_configuration(),
        )
        task.train(dataset, output_model, train_parameters)

        assert len(training_progress_curve) > 0
        assert np.all(training_progress_curve[1:] >= training_progress_curve[:-1])

    @e2e_pytest_api
    def test_inference_progress_tracking(self):
        hyper_parameters, model_template = self.setup_configurable_parameters(DEFAULT_ACTION_TEMPLATE_DIR, num_iters=10)
        action_environment, dataset = self.init_environment(hyper_parameters, model_template)

        task = ActionInferenceTask(task_environment=action_environment)
        print("Task initialized, model inference starts.")
        inference_progress_curve = []

        def progress_callback(progress: int):
            assert isinstance(progress, int)
            inference_progress_curve.append(progress)

        inference_parameters = InferenceParameters()
        inference_parameters.update_progress = progress_callback
        task.infer(dataset, inference_parameters)

        assert len(inference_progress_curve) > 0
        assert np.all(inference_progress_curve[1:] >= inference_progress_curve[:-1])

    @e2e_pytest_api
    def test_inference_task(self):
        # FIXME CVS-103071 will handle this
        # Prepare pretrained weights
        hyper_parameters, model_template = self.setup_configurable_parameters(DEFAULT_ACTION_TEMPLATE_DIR, num_iters=2)
        action_environment, dataset = self.init_environment(hyper_parameters, model_template)
        # val_dataset = dataset.get_subset(Subset.VALIDATION)

        train_task = ActionTrainTask(task_environment=action_environment)

        training_progress_curve = []

        def progress_callback(progress: float, score: Optional[float] = None):
            training_progress_curve.append(progress)

        train_parameters = TrainParameters()
        train_parameters.update_progress = progress_callback
        trained_model = ModelEntity(
            dataset,
            action_environment.get_model_configuration(),
        )
        train_task.train(dataset, trained_model, train_parameters)
        # performance_after_train = task_eval(train_task, trained_model, val_dataset)

        # Create InferenceTask
        # action_environment.model = trained_model
        # inference_task = ActionInferenceTask(task_environment=action_environment)

        # performance_after_load = task_eval(inference_task, trained_model, val_dataset)

        # FIXME CVS-103071 will handle this
        # assert performance_after_train == performance_after_load

        # Export
        # CVS-102941 ONNX export of action detection model keeps failed
        # exported_model = ModelEntity(dataset, action_environment.get_model_configuration())
        # inference_task.export(ExportType.OPENVINO, exported_model)



"""This module contains utility functions used within the configuration helper module."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import json
import os
from enum import Enum
from typing import Any, List, Tuple, Type, Union

import yaml
from omegaconf import DictConfig, OmegaConf

from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.configuration.enums.utils import get_enum_names
from otx.api.entities.id import ID

from .config_element_mapping import (
    GroupElementMapping,
    PrimitiveElementMapping,
    RuleElementMapping,
)
from .convert import convert


def _search_in_config_dict_inner(
    config_dict: dict,
    key_to_search: str,
    prior_keys: List[str] = None,
    results: List[Tuple[Any, List[str]]] = None,
) -> List[Tuple[Any, List[str]]]:
    """Helper function for the `search_in_config_dict` function defined below.

    Args:
        config_dict: dict to search in
        key_to_search: dict key to look for
        prior_keys: List of prior keys leading to the key_to_search.
        results: List of previously found results

    Returns:
        List of (value_at_key_to_search, key_path_to_key_to_search)
        tuples, representing each occurrence of key_to_search within
        config_dict
    """
    if prior_keys is None:
        prior_keys = []
    if results is None:
        results = []
    if isinstance(config_dict, List):
        dict_to_search_in = dict(zip(range(len(config_dict)), config_dict))
    else:
        dict_to_search_in = config_dict
    if not (issubclass(type(dict_to_search_in), dict) or isinstance(dict_to_search_in, DictConfig)):
        return results
    for key, value in dict_to_search_in.items():
        current_key_path = prior_keys + [key]
        if key == key_to_search:
            results.append((value, prior_keys))
        _search_in_config_dict_inner(value, key_to_search, current_key_path, results)
    return results


def search_in_config_dict(config_dict: dict, key_to_search: str) -> List[Tuple[Any, List[str]]]:
    """Recursively searches a config_dict for all instances of key_to_search and returns the key path to them.

    Args:
        config_dict: dict to search in
        key_to_search: dict key to look for

    Returns:
        List of (value_at_key_to_search, key_path_to_key_to_search)
        tuples, representing each occurrence of key_to_search within
        config_dict
    """
    return _search_in_config_dict_inner(config_dict, key_to_search=key_to_search)


def input_to_config_dict(input_config: Union[str, DictConfig, dict], check_config_type: bool = True) -> dict:
    """Validate input configuration.

    Takes an input_config which can be a string, filepath, dict or DictConfig and performs basic validation that it
    can be converted into a configuration.

    Args:
        input_config: String, filepath, dict or DictConfig describing a
            configuration
        check_config_type: True to check that the input has a proper
            `type` attribute in order to be converted into a
            ConfigurableParameters object. False to disable this check.
            Defaults to True.

    Returns:
        dictionary or DictConfig
    """
    valid_types = (
        get_enum_names(PrimitiveElementMapping)
        + get_enum_names(RuleElementMapping)
        + get_enum_names(GroupElementMapping)
    )

    if isinstance(input_config, str):
        if os.path.exists(input_config):
            with open(input_config, "r", encoding="UTF-8") as file:
                result = yaml.safe_load(file)
        else:
            result = yaml.safe_load(input_config)
    elif issubclass(type(input_config), dict):
        result = input_config
    elif isinstance(input_config, DictConfig):
        result = OmegaConf.to_container(input_config)
    else:
        raise ValueError('Invalid input_config type! Valid types are "str", "DictConfig", "dict".')

    if check_config_type:
        config_type = str(result.get("type", None))

        if config_type is None:
            raise ValueError(
                f"Input cannot be converted to a valid configuration. No "
                f"configuration type was found in the following input: {input_config}"
            )
        if config_type not in valid_types:
            raise ValueError(
                f"Invalid configuration element type: {config_type} found in input. "
                f"Unable to parse input. Supported configuration element types are:"
                f" {valid_types}"
            )
    return result


def deserialize_enum_value(value: Union[str, Enum], enum_type: Type[Enum]):
    """Deserializes a value to an instance of a certain Enum.

    This checks whether the `value` passed is already an instance of the target Enum, in which case this function just
    returns the input `value`. If value is a string, this function returns the corresponding instance of the Enum
    passed in `enum_type`.

    Args:
        value: value to deserialize
        enum_type: class (should be a subclass of Enum) that the name
            belongs to

    Returns:
        instance of `enum_type`.`value`
    """
    if isinstance(value, enum_type):
        instance = value
    elif isinstance(value, str):
        try:
            instance = enum_type[value.upper()]
        except KeyError:
            try:
                instance = enum_type(value)  # type: ignore
            except ValueError as error:
                raise KeyError(f"Value `{value}` cannot be converted to an instance of " f"`{enum_type}`") from error
    else:
        raise ValueError(
            f"Invalid input data type, {type(value)} cannot be converted to an instance " f"of {enum_type}."
        )
    return instance


def ids_to_strings(config_dict: dict) -> dict:
    """Converts ID's in the `config_dict` to their string representation.

    Args:
        config_dict: Dictionary in which to replace the ID's by strings

    Returns:
        Updated config_dict dictionary
    """
    for key, value in config_dict.items():
        if isinstance(value, ID):
            config_dict[key] = str(value)
    return config_dict


def config_to_bytes(config: ConfigurableParameters) -> bytes:
    """Converts ConfigurableParameters to bytes.

    Args:
        config: configurable parameters

    Retruns:
        JSON in bytes
    """
    config_dict = convert(config, dict, enum_to_str=True)
    return json.dumps(config_dict, indent=4).encode()



_base_ = [
    "../_base_/default.py",
    "../_base_/logs/tensorboard_logger.py",
    "../_base_/optimizers/sgd.py",
    "../_base_/runners/epoch_runner_cancel.py",
    "../_base_/schedules/plateau.py",
]

optimizer = dict(
    lr=0.001,
    momentum=0.9,
    weight_decay=0.0001,
)

lr_config = dict(
    policy="ReduceLROnPlateau",
    metric="mDice",
    patience=5,
    iteration_patience=0,
    interval=1,
    min_lr=0.000001,
    warmup="linear",
    warmup_iters=80,
    warmup_ratio=1.0 / 3,
)

evaluation = dict(
    interval=1,
    metric=["mIoU", "mDice"],
)

custom_hooks = [
    dict(
        type="LazyEarlyStoppingHook",
        patience=8,
        iteration_patience=0,
        metric="mDice",
        interval=1,
        priority=75,
        start=1,
    ),
]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import math
import random

import numpy as np
from torch.utils.data.sampler import Sampler

from otx.mpa.modules.utils.task_adapt import unwrap_dataset


class ClsIncrSampler(Sampler):
    """Sampler for Class-Incremental Task
    This sampler is a sampler that creates an effective batch
    For default setting,
    the square root of (number of old data/number of new data) is used as the ratio of old data
    In effective mode,
    the ratio of old and new data is used as 1:1

    Args:
        dataset (Dataset): A built-up dataset
        samples_per_gpu (int): batch size of Sampling
        efficient_mode (bool): Flag about using efficient mode
    """

    def __init__(self, dataset, samples_per_gpu, efficient_mode=False, num_replicas=1, rank=0, drop_last=False):
        self.samples_per_gpu = samples_per_gpu
        self.num_replicas = num_replicas
        self.rank = rank
        self.drop_last = drop_last

        # Dataset Wrapping remove & repeat for RepeatDataset
        self.dataset, self.repeat = unwrap_dataset(dataset)

        if hasattr(self.dataset, "img_indices"):
            self.new_indices = self.dataset.img_indices["new"]
            self.old_indices = self.dataset.img_indices["old"]
        else:
            raise TypeError(f"{self.dataset} type does not have img_indices")

        if not len(self.new_indices) > 0:
            self.new_indices = self.old_indices
            self.old_indices = []

        old_new_ratio = np.sqrt(len(self.old_indices) / len(self.new_indices))

        if efficient_mode:
            self.data_length = int(len(self.new_indices) * (1 + old_new_ratio))
            self.old_new_ratio = 1
        else:
            self.data_length = len(self.dataset)
            self.old_new_ratio = int(old_new_ratio)

        self.num_samples = self._calcuate_num_samples()

    def _calcuate_num_samples(self):
        num_samples = self.repeat * (1 + self.old_new_ratio) * int(self.data_length / (1 + self.old_new_ratio))

        if not self.drop_last:
            num_samples += (
                int(np.ceil(self.data_length * self.repeat / self.samples_per_gpu)) * self.samples_per_gpu - num_samples
            )

        if self.num_replicas > 1:
            # If the dataset length is evenly divisible by # of replicas, then there
            # is no need to drop any data, since the dataset will be split equally.
            if self.drop_last and num_samples % self.num_replicas != 0:  # type: ignore
                # Split to nearest available length that is evenly divisible.
                # This is to ensure each rank receives the same amount of data when
                # using this Sampler.
                num_samples = math.ceil(
                    # `type:ignore` is required because Dataset cannot provide a default __len__
                    # see NOTE in pytorch/torch/utils/data/sampler.py
                    (num_samples - self.num_replicas)
                    / self.num_replicas  # type: ignore
                )
            else:
                num_samples = math.ceil(num_samples / self.num_replicas)  # type: ignore
            self.total_size = num_samples * self.num_replicas

        return num_samples

    def __iter__(self):
        indices = []
        for _ in range(self.repeat):
            for i in range(int(self.data_length / (1 + self.old_new_ratio))):
                indice = np.concatenate(
                    [np.random.choice(self.new_indices, 1), np.random.choice(self.old_indices, self.old_new_ratio)]
                )
                indices.append(indice)

        indices = np.concatenate(indices)
        if not self.drop_last:
            num_extra = int(
                np.ceil(self.data_length * self.repeat / self.samples_per_gpu)
            ) * self.samples_per_gpu - len(indices)
            indices = np.concatenate([indices, np.random.choice(indices, num_extra)])
        indices = indices.astype(np.int64).tolist()

        if self.num_replicas > 1:
            if not self.drop_last:
                # add extra samples to make it evenly divisible
                padding_size = self.total_size - len(indices)
                # add extra samples to make it evenly divisible
                if padding_size <= len(indices):
                    indices += indices[:padding_size]
                else:
                    indices += (indices * math.ceil(padding_size / len(indices)))[:padding_size]
            else:
                # remove tail of data to make it evenly divisible.
                indices = indices[: self.total_size]
            assert len(indices) == self.total_size

            # shuffle before distributing indices
            random.shuffle(indices)

            # subsample
            indices = indices[self.rank : self.total_size : self.num_replicas]
            assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self):
        return self.num_samples



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import math

from mmcv.parallel import is_module_wrapper
from mmcv.runner import HOOKS, Hook


@HOOKS.register_module()
class SemiSLClsHook(Hook):
    """Hook for SemiSL for classification

    This hook includes unlabeled warm-up loss coefficient (default: True):
        unlabeled_coef = (0.5 - cos(min(pi, 2 * pi * k) / K)) / 2
        k: current step, K: total steps
    Also, this hook adds semi-sl-related data to the log (unlabeled_coef, pseudo_label)

    Args:
        total_steps (int): total steps for training (iteration)
            Raise the coefficient from 0 to 1 during half the duration of total_steps
            default: 0, use runner.max_iters
        unlabeled_warmup (boolean): enable unlabeled warm-up loss coefficient
            If False, Semi-SL uses 1 as unlabeled loss coefficient
    """

    def __init__(self, total_steps=0, unlabeled_warmup=True, **kwargs):
        self.unlabeled_warmup = unlabeled_warmup
        self.total_steps = total_steps
        self.current_step, self.unlabeled_coef = 0, 0
        self.num_pseudo_label = 0

    def before_train_iter(self, runner):
        # Calculate the unlabeled warm-up loss coefficient before training iteration
        if self.unlabeled_warmup and self.unlabeled_coef < 1.0:
            if self.total_steps == 0:
                self.total_steps = runner.max_iters
            self.unlabeled_coef = 0.5 * (
                1 - math.cos(min(math.pi, (2 * math.pi * self.current_step) / self.total_steps))
            )
            model = self._get_model(runner)
            model.head.unlabeled_coef = self.unlabeled_coef
        self.current_step += 1

    def after_train_iter(self, runner):
        model = self._get_model(runner)
        # Add the number of pseudo-labels currently selected from iteration
        self.num_pseudo_label += int(model.head.num_pseudo_label)

    def after_epoch(self, runner):
        # Add data related to Semi-SL to the log
        if self.unlabeled_warmup:
            runner.log_buffer.output.update({"unlabeled_coef": round(self.unlabeled_coef, 4)})
        runner.log_buffer.output.update({"pseudo_label": self.num_pseudo_label})
        self.num_pseudo_label = 0

    def _get_model(self, runner):
        model = runner.model
        if is_module_wrapper(model):
            model = model.module
        return model



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import random
from pathlib import Path

import cv2 as cv
import numpy as np

from otx.api.configuration.helper import create
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelGroup, LabelGroupType, LabelSchemaEntity
from otx.api.entities.model_template import parse_model_template
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.mpa.utils.config_utils import MPAConfig

DEFAULT_CLS_TEMPLATE_DIR = Path("otx") / "algorithms" / "classification" / "configs" / "mobilenet_v3_large_1_cls_incr"
DEFAULT_CLS_TEMPLATE = DEFAULT_CLS_TEMPLATE_DIR / "template.yaml"


def generate_label_schema(not_empty_labels, multilabel=False, hierarchical=False):
    assert len(not_empty_labels) > 1

    label_schema = LabelSchemaEntity()
    if multilabel:
        emptylabel = LabelEntity(name="Empty label", is_empty=True, domain=Domain.CLASSIFICATION)
        empty_group = LabelGroup(name="empty", labels=[emptylabel], group_type=LabelGroupType.EMPTY_LABEL)
        for label in not_empty_labels:
            label_schema.add_group(
                LabelGroup(
                    name=label.name,
                    labels=[label],
                    group_type=LabelGroupType.EXCLUSIVE,
                )
            )
        label_schema.add_group(empty_group)
    elif hierarchical:
        single_label_classes = ["b", "g", "r"]
        multi_label_classes = ["w", "p"]
        emptylabel = LabelEntity(name="Empty label", is_empty=True, domain=Domain.CLASSIFICATION)
        empty_group = LabelGroup(name="empty", labels=[emptylabel], group_type=LabelGroupType.EMPTY_LABEL)
        single_labels = []
        for label in not_empty_labels:
            if label.name in multi_label_classes:
                label_schema.add_group(
                    LabelGroup(
                        name=label.name,
                        labels=[label],
                        group_type=LabelGroupType.EXCLUSIVE,
                    )
                )
                if empty_group not in label_schema.get_groups(include_empty=True):
                    label_schema.add_group(empty_group)
            elif label.name in single_label_classes:
                single_labels.append(label)
        if single_labels:
            single_label_group = LabelGroup(
                name="labels",
                labels=single_labels,
                group_type=LabelGroupType.EXCLUSIVE,
            )
            label_schema.add_group(single_label_group)
    else:
        main_group = LabelGroup(
            name="labels",
            labels=not_empty_labels,
            group_type=LabelGroupType.EXCLUSIVE,
        )
        label_schema.add_group(main_group)
    return label_schema


def generate_cls_dataset(hierarchical=False, number_of_images=10):
    resolution = (224, 224)
    if hierarchical:
        colors = [(0, 255, 0), (0, 0, 255), (255, 0, 0), (0, 0, 0), (230, 230, 250)]
        cls_names = ["b", "g", "r", "w", "p"]
        texts = ["Blue", "Green", "Red", "White", "Purple"]
    else:
        colors = [(0, 255, 0), (0, 0, 255)]
        cls_names = ["b", "g"]
        texts = ["Blue", "Green"]
    env_labels = [
        LabelEntity(name=name, domain=Domain.CLASSIFICATION, is_empty=False, id=ID(i))
        for i, name in enumerate(cls_names)
    ]

    items = []

    for _ in range(0, number_of_images):
        for j, lbl in enumerate(env_labels):
            class_img = np.zeros((*resolution, 3), dtype=np.uint8)
            class_img[:] = colors[j]
            class_img = cv.putText(
                class_img,
                texts[j],
                (50, 50),
                cv.FONT_HERSHEY_SIMPLEX,
                0.8 + j * 0.2,
                colors[j - 1],
                2,
                cv.LINE_AA,
            )

            image = Image(data=class_img)
            labels = [ScoredLabel(label=lbl, probability=1.0)]
            shapes = [Annotation(Rectangle.generate_full_box(), labels)]
            annotation_scene = AnnotationSceneEntity(kind=AnnotationSceneKind.ANNOTATION, annotations=shapes)
            items.append(DatasetItemEntity(media=image, annotation_scene=annotation_scene))

    rng = random.Random()
    rng.seed(100)
    rng.shuffle(items)
    for i, _ in enumerate(items):
        subset_region = i / number_of_images
        if subset_region >= 0.9:
            subset = Subset.TESTING
        elif subset_region >= 0.6:
            subset = Subset.VALIDATION
        else:
            subset = Subset.TRAINING
        items[i].subset = subset

    dataset = DatasetEntity(items)
    return dataset


def init_environment(params, model_template, multilabel=False, hierarchical=False, number_of_images=10):
    dataset = generate_cls_dataset(hierarchical, number_of_images)
    labels_schema = generate_label_schema(dataset.get_labels(), multilabel=multilabel, hierarchical=hierarchical)
    environment = TaskEnvironment(
        model=None,
        hyper_parameters=params,
        label_schema=labels_schema,
        model_template=model_template,
    )
    return environment, dataset


def setup_configurable_parameters(template_dir, num_iters=10):
    model_template = parse_model_template(str(template_dir))
    hyper_parameters = create(model_template.hyper_parameters.data)
    hyper_parameters.learning_parameters.num_iters = num_iters

    return hyper_parameters, model_template


def setup_mpa_task_parameters(task_type, create_val=False, create_test=False):
    if task_type == "semisl":
        recipie_path = "otx/recipes/stages/classification/semisl.yaml"
    elif task_type == "incremental":
        recipie_path = "otx/recipes/stages/classification/incremental.yaml"
    recipie_cfg = MPAConfig.fromfile(recipie_path)
    model_cfg = MPAConfig.fromfile(DEFAULT_CLS_TEMPLATE_DIR / "model.py")
    model_cfg.model.multilabel = False
    model_cfg.model.hierarchical = False
    data_cfg = MPAConfig.fromfile(DEFAULT_CLS_TEMPLATE_DIR / "data_pipeline.py")
    data_cfg.data.train.data_dir = "tests/assets/imagenet_dataset"
    if create_val:
        data_cfg.data.val.data_dir = "tests/assets/imagenet_dataset"
    else:
        data_cfg.data.val = None
    if create_test:
        data_cfg.data.test.data_dir = "tests/assets/imagenet_dataset"
    else:
        data_cfg.data.test = None
    dummy_dataset = generate_cls_dataset(number_of_images=1)
    data_cfg.data.train.otx_dataset = dummy_dataset
    data_cfg.data.train.labels = dummy_dataset.get_labels()
    data_cfg.data.train.data_classes = ["label_0", "label_1"]
    data_cfg.data.train.new_classes = ["label_0", "label_1", "label_3"]

    return model_cfg, data_cfg, recipie_cfg



"""Collection of Action detection evaluiation utils.."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import logging
import time
from collections import defaultdict

import numpy as np
from mmaction.core.evaluation.ava_evaluation import (
    object_detection_evaluation as mm_det_eval,
)
from mmaction.core.evaluation.ava_evaluation import standard_fields
from mmaction.core.evaluation.ava_utils import print_time, read_exclusions


# pylint: disable=too-many-locals, too-many-branches
def det_eval(predictions, result_type, labels, video_infos, exclude_file, verbose=True, custom_classes=None):
    """Evaluation method for AVA Dataset."""

    assert result_type in ["mAP"]

    start = time.time()
    categories, class_whitelist = _read_labelmap(labels)
    if custom_classes is not None:
        custom_classes = custom_classes[1:]
        assert set(custom_classes).issubset(set(class_whitelist))
        class_whitelist = custom_classes
        categories = [cat for cat in categories if cat["id"] in custom_classes]

    # loading gt, do not need gt score
    gt_boxes, gt_labels = _load_gt(video_infos)
    if verbose:
        print_time("Reading detection results", start)

    if exclude_file is not None:
        with open(exclude_file, encoding="utf-8") as ex_file:
            excluded_keys = read_exclusions(ex_file)
    else:
        excluded_keys = []

    start = time.time()
    boxes, labels, scores = predictions
    if verbose:
        print_time("Reading detection results", start)

    # Evaluation for mAP
    pascal_evaluator = mm_det_eval.PascalDetectionEvaluator(categories)

    start = time.time()
    for image_key in gt_boxes:
        if verbose and image_key in excluded_keys:
            logging.info("Found excluded timestamp in detections: %s. It will be ignored.", image_key)
            continue
        pascal_evaluator.add_single_ground_truth_image_info(
            image_key,
            {
                standard_fields.InputDataFields.groundtruth_boxes: np.array(gt_boxes[image_key], dtype=float),
                standard_fields.InputDataFields.groundtruth_classes: np.array(gt_labels[image_key], dtype=int),
            },
        )
    if verbose:
        print_time("Convert groundtruth", start)

    start = time.time()
    for image_key in boxes:
        if verbose and image_key in excluded_keys:
            logging.info("Found excluded timestamp in detections: %s. It will be ignored.", image_key)
            continue
        pascal_evaluator.add_single_detected_image_info(
            image_key,
            {
                standard_fields.DetectionResultFields.detection_boxes: np.array(boxes[image_key], dtype=float),
                standard_fields.DetectionResultFields.detection_classes: np.array(labels[image_key], dtype=int),
                standard_fields.DetectionResultFields.detection_scores: np.array(scores[image_key], dtype=float),
            },
        )
    if verbose:
        print_time("convert detections", start)

    start = time.time()
    metrics = pascal_evaluator.evaluate()
    if verbose:
        print_time("run_evaluator", start)
    for display_name, value in metrics.items():
        print(f"{display_name}=\t{value}")
    return {display_name: value for display_name, value in metrics.items() if "ByCategory" not in display_name}


def _read_labelmap(labels):
    """Generate label map from LabelEntity."""
    labelmap = []
    class_ids = set()
    for label in labels:
        labelmap.append({"id": int(label.id), "name": str(label.name)})
        class_ids.add(int(label.id))
    return labelmap, class_ids


def _load_gt(video_infos):
    """Generate ground truth information from video_infos."""
    boxes = defaultdict(list)
    labels = defaultdict(list)
    for video_info in video_infos:
        img_key = video_info["img_key"]
        gt_bboxes = video_info["gt_bboxes"]
        gt_labels = video_info["gt_labels"]
        for gt_label, gt_bbox in zip(gt_labels, gt_bboxes):
            for idx, val in enumerate(gt_label):
                if val == 1:
                    boxes[img_key].append(gt_bbox)
                    labels[img_key].append(idx)
    return boxes, labels



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest

from otx.algorithms.detection.adapters.mmdet.data.dataset import (
    OTXDetDataset,
    get_annotation_mmdet_format,
)
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)


def label_entity():
    return LabelEntity(name="test label", domain=Domain.DETECTION)


def dataset_item():
    image = Image(data=np.random.randint(low=0, high=255, size=(10, 16, 3)))
    annotation = Annotation(shape=Rectangle.generate_full_box(), labels=[ScoredLabel(label_entity())])
    annotation_scene = AnnotationSceneEntity(annotations=[annotation], kind=AnnotationSceneKind.ANNOTATION)
    return DatasetItemEntity(media=image, annotation_scene=annotation_scene)


class TestMMDatasetFunctionsInputParamsValidation:
    @e2e_pytest_unit
    def test_get_annotation_mmdet_format_input_params_validation(self):
        """
        <b>Description:</b>
        Check "get_annotation_mmdet_format" function input parameters validation

        <b>Input data:</b>
        "get_annotation_mmdet_format" function unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "get_annotation_mmdet_format" function
        """
        label = label_entity()
        correct_values_dict = {
            "dataset_item": dataset_item(),
            "labels": [label],
            "domain": Domain.DETECTION,
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "dataset_item" parameter
            ("dataset_item", unexpected_int),
            # Unexpected integer is specified as "labels" parameter
            ("labels", unexpected_int),
            # Unexpected integer is specified as nested label
            ("labels", [label, unexpected_int]),
            # Unexpected integer is specified as "domain" parameter
            ("domain", unexpected_int),
            # Unexpected string is specified as "min_size" parameter
            ("min_size", "unexpected string"),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=get_annotation_mmdet_format,
        )


class TestOTXDetDatasetInputParamsValidation:
    @staticmethod
    def dataset():
        pipeline = [{"type": "LoadImageFromFile", "to_float32": True}]
        return OTXDetDataset(
            otx_dataset=DatasetEntity(),
            labels=[label_entity()],
            pipeline=pipeline,
            test_mode=True,
            domain=Domain.DETECTION,
        )

    @e2e_pytest_unit
    def test_otx_dataset_init_params_validation(self):
        """
        <b>Description:</b>
        Check OTXDetDataset object initialization parameters validation

        <b>Input data:</b>
        OTXDetDataset object initialization parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        OTXDetDataset object initialization parameter
        """
        label = label_entity()

        correct_values_dict = {
            "otx_dataset": DatasetEntity(),
            "labels": [label],
            "pipeline": [{"type": "LoadImageFromFile", "to_float32": True}],
            "domain": Domain.DETECTION,
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "otx_dataset" parameter
            ("otx_dataset", unexpected_str),
            # Unexpected string is specified as "labels" parameter
            ("labels", unexpected_str),
            # Unexpected string is specified as nested label
            ("labels", [label, unexpected_str]),
            # Unexpected integer is specified as "pipeline" parameter
            ("pipeline", 1),
            # Unexpected string is specified as nested pipeline
            ("pipeline", [{"config": 1}, unexpected_str]),
            # Unexpected string is specified as "domain" parameter
            ("domain", unexpected_str),
            # Unexpected string is specified as "test_mode" parameter
            ("test_mode", unexpected_str),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=OTXDetDataset,
        )

    @e2e_pytest_unit
    def test_otx_dataset_prepare_train_img_params_validation(self):
        """
        <b>Description:</b>
        Check OTXDetDataset object "prepare_train_img" method input parameters validation

        <b>Input data:</b>
        OTXDetDataset object, "idx" non-integer type parameter

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "prepare_train_img" method
        """
        dataset = self.dataset()
        with pytest.raises(ValueError):
            dataset.prepare_train_img(idx="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_otx_dataset_prepare_test_img_params_validation(self):
        """
        <b>Description:</b>
        Check OTXDetDataset object "prepare_test_img" method input parameters validation

        <b>Input data:</b>
        OTXDetDataset object, "idx" non-integer type parameter

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "prepare_test_img" method
        """
        dataset = self.dataset()
        with pytest.raises(ValueError):
            dataset.prepare_test_img(idx="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_otx_dataset_pre_pipeline_params_validation(self):
        """
        <b>Description:</b>
        Check OTXDetDataset object "pre_pipeline" method input parameters validation

        <b>Input data:</b>
        OTXDetDataset object, "results" unexpected type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "pre_pipeline" method
        """
        dataset = self.dataset()
        unexpected_int = 1
        for unexpected_value in [
            # Unexpected integer is specified as "results" parameter
            unexpected_int,
            # Unexpected integer is specified as "results" dictionary key
            {"result_1": "some results", unexpected_int: "unexpected results"},
        ]:
            with pytest.raises(ValueError):
                dataset.pre_pipeline(results=unexpected_value)

    @e2e_pytest_unit
    def test_otx_dataset_get_ann_info_params_validation(self):
        """
        <b>Description:</b>
        Check OTXDetDataset object "get_ann_info" method input parameters validation

        <b>Input data:</b>
        OTXDetDataset object, "idx" non-integer type parameter

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "get_ann_info" method
        """
        dataset = self.dataset()
        with pytest.raises(ValueError):
            dataset.get_ann_info(idx="unexpected string")  # type: ignore



"""Unit Test for otx.algorithms.action.adapters.openvino.dataloader."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from copy import deepcopy
from typing import List, Optional

import pytest

from otx.algorithms.action.adapters.openvino.dataloader import (
    ActionOVClsDataLoader,
    ActionOVDemoDataLoader,
    ActionOVDetDataLoader,
    _is_multi_video,
    get_ovdataloader,
)
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.metadata import MetadataItemEntity, VideoMetadata
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.action.test_helpers import (
    generate_action_cls_otx_dataset,
    generate_action_det_otx_dataset,
    generate_labels,
)


class MockDatasetEntity(DatasetEntity):
    """Mock class for DatasetEntity."""

    def __init__(self):
        pass


class MockDatasetItemEntity(DatasetItemEntity):
    """Mock class for DatasetItemEntity"""

    def __init__(
        self,
        metadata: List[MetadataItemEntity],
        annotation_scene: Optional[AnnotationSceneEntity] = None,
    ):
        self.__metadata = metadata
        self.__annotation_scene: AnnotationSceneEntity = annotation_scene

    def get_metadata(self) -> List[MetadataItemEntity]:
        return self.__metadata


@e2e_pytest_unit
def test_get_ovdataloader(mocker) -> None:
    """Test get_ovdataloader function.
    <Steps>
        1. Check ovdataloader type when function get ACTION_CLASSIFIACTION as task_type
        2. Check ovdataloader type when function get ACTION_DETECTION as task_type
        3. Check ovdataloader type when function get single video dataset
    """

    class MockActionOVClsDataLoader(ActionOVClsDataLoader):
        """Mock class for ActionOVClsDataLoader."""

        def __init__(self, dataset: DatasetEntity, clip_len: int, width: int, height: int):
            pass

    class MockActionOVDetDataLoader(ActionOVDetDataLoader):
        """Mock class for ActionOVDetDataLoader."""

        def __init__(self, dataset: DatasetEntity, clip_len: int, width: int, height: int):
            pass

    mocker.patch("otx.algorithms.action.adapters.openvino.dataloader._is_multi_video", return_value=True)
    mocker.patch(
        "otx.algorithms.action.adapters.openvino.dataloader.ActionOVClsDataLoader",
        side_effect=MockActionOVClsDataLoader,
    )
    mocker.patch(
        "otx.algorithms.action.adapters.openvino.dataloader.ActionOVDetDataLoader",
        side_effect=MockActionOVDetDataLoader,
    )

    _task_type = "ACTION_CLASSIFICATION"
    out = get_ovdataloader(MockDatasetEntity(), _task_type, 8, 256, 256)
    assert isinstance(out, ActionOVClsDataLoader)

    _task_type = "ACTION_DETECTION"
    out = get_ovdataloader(MockDatasetEntity(), _task_type, 8, 256, 256)
    assert isinstance(out, ActionOVDetDataLoader)

    _task_type = "ACTION_SEGMENTATION"
    with pytest.raises(NotImplementedError):
        out = get_ovdataloader(MockDatasetEntity, _task_type, 8, 256, 256)

    mocker.patch("otx.algorithms.action.adapters.openvino.dataloader._is_multi_video", return_value=False)

    _task_type = "ACTION_DETECTION"
    out = get_ovdataloader(MockDatasetEntity(), _task_type, 8, 256, 256)
    assert isinstance(out, ActionOVDemoDataLoader)


@e2e_pytest_unit
def test_is_multi_video() -> None:
    """Test _is_multi_video function.

    <Steps>
        1. Check return value: bool when function get single video DatasetEntity
        2. Check return value: bool when function get multi video DatasetEntity
    """

    items: List[DatasetItemEntity] = []
    items.append(MockDatasetItemEntity(metadata=[MetadataItemEntity(data=VideoMetadata("2", 0, False))]))
    items.append(MockDatasetItemEntity(metadata=[MetadataItemEntity(data=VideoMetadata("2", 1, False))]))
    items.append(MockDatasetItemEntity(metadata=[MetadataItemEntity(data=VideoMetadata("2", 2, False))]))

    dataset = DatasetEntity(items)
    assert _is_multi_video(dataset) is False

    items.append(MockDatasetItemEntity(metadata=[MetadataItemEntity(data=VideoMetadata("1", 0, False))]))
    items.append(MockDatasetItemEntity(metadata=[MetadataItemEntity(data=VideoMetadata("1", 1, False))]))
    items.append(MockDatasetItemEntity(metadata=[MetadataItemEntity(data=VideoMetadata("1", 2, False))]))

    dataset = DatasetEntity(items)
    assert _is_multi_video(dataset) is True


class TestActionOVDemoDataLoader:
    """Test ActionOVDemoDataLoader class.

    1. Initialize ActionOVDemoDataLoader and check its length
    2. Test __getitem__ function
    <Steps>
        1. Create ActionOVDemoDataLoader
        2. Sample first item from ActionOVDemoDataLoader
        3. The item's frame indices should be [0, 0, 0, 0, 0, 2, 4, 6]
    3. Test add_prediction function
    <Steps>
        1. Create sample prediction
        2. Check whether empty annotation changed to sample prediction
    """

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.data_len = 40
        self.labels = generate_labels(1, Domain.ACTION_CLASSIFICATION)
        self.dataset = generate_action_cls_otx_dataset(1, self.data_len, self.labels)
        self.dataloader = ActionOVClsDataLoader(self.dataset, 8, 256, 256)

    @e2e_pytest_unit
    def test_len(self) -> None:
        """Test initialization and __len__ function."""

        dataloader = ActionOVDemoDataLoader(self.dataset, "ACTION_CLASSIFICATION", 8, 256, 256)
        assert len(dataloader) == self.data_len

    @e2e_pytest_unit
    def test_getitem(self) -> None:
        """Test __getitem__ function."""

        dataloader = ActionOVDemoDataLoader(self.dataset, "ACTION_CLASSIFICATION", 8, 256, 256)
        outs = dataloader[0]
        frame_indices: List[int] = []
        for out in outs:
            frame_idx = out.get_metadata()[0].data.frame_idx
            frame_indices.append(frame_idx)
        assert frame_indices == [0, 0, 0, 0, 0, 2, 4, 6]

    @e2e_pytest_unit
    def test_add_prediction(self) -> None:
        """Test add_prediction function."""
        prediction = AnnotationSceneEntity(
            annotations=[
                Annotation(
                    Rectangle.generate_full_box(),
                    [ScoredLabel(LabelEntity("2", Domain.ACTION_CLASSIFICATION, id=ID(1)))],
                )
            ],
            kind=AnnotationSceneKind.ANNOTATION,
        )

        dataloader = ActionOVDemoDataLoader(self.dataset, "ACTION_CLASSIFICATION", 8, 256, 256)
        items = self.dataset.with_empty_annotations()._items
        dataloader.add_prediction(items, prediction)
        assert len(items[0].get_annotations()) == 0
        assert len(items[len(items) // 2].get_annotations()) >= 1

        dataloader = ActionOVDemoDataLoader(self.dataset, "ACTION_DETECTION", 8, 256, 256)
        items = self.dataset.with_empty_annotations()._items
        dataloader.add_prediction(items, prediction)
        assert len(items[0].get_annotations()) == 0
        assert len(items[len(items) // 2].get_annotations()) >= 1


class TestActionOVClsDataLoader:
    """Test class for ActionOVClsDataLoader.

    1. Test initialization
    <Steps>
        1. Check self.dataloader's length. It should be 1 because all dataset item have same video_id
        2. Check self.dataloader.dataset's length. It should be self.data_len(40)
    2. Test __getitem__
    <Steps>
        1. Check len(output) == clip_len(8)
        2. Check frame_indices == [4, 8, 12, 16, 20, 24, 28, 32]. It comes from setting indices rule.
    3. Test add_prediciton
    <Steps>
        1. Check self.dataset.get_labels(). It should be 1, because only one label is added to empty dataset.
        2. Check self.dataset's label's id is 2.
    """

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.data_len = 40
        self.labels = generate_labels(1, Domain.ACTION_CLASSIFICATION)
        self.dataset = generate_action_cls_otx_dataset(1, self.data_len, self.labels)
        self.dataloader = ActionOVClsDataLoader(self.dataset, 8, 256, 256)

    @e2e_pytest_unit
    def test_init(self) -> None:
        """Test __init__ function."""

        assert len(self.dataloader) == 1
        assert len(self.dataloader.dataset[0]) == self.data_len

    @e2e_pytest_unit
    def test_getitem(self) -> None:
        """Test __getitem__ function."""

        outs = self.dataloader[0]
        frame_indices = []
        for out in outs:
            frame_idx = out.get_metadata()[0].data.frame_idx
            frame_indices.append(frame_idx)
        assert len(outs) == 8
        assert frame_indices == [4, 8, 12, 16, 20, 24, 28, 32]

    @e2e_pytest_unit
    def test_add_prediction(self) -> None:
        """Test add_prediciton function."""

        prediction = AnnotationSceneEntity(
            annotations=[
                Annotation(
                    Rectangle.generate_full_box(),
                    [ScoredLabel(LabelEntity("2", Domain.ACTION_CLASSIFICATION, id=ID(2)))],
                )
            ],
            kind=AnnotationSceneKind.ANNOTATION,
        )
        items = deepcopy(self.dataset._items)
        self.dataset = self.dataset.with_empty_annotations()
        self.dataloader.add_prediction(self.dataset, items, prediction)
        assert len(self.dataset.get_labels()) == 1
        assert int(self.dataset.get_labels()[0].id) == 2


class TestActionOVDetDataLoader:
    """Test class for ActionOVDetDataLoader.

    1. Test initialization
    <Steps>
        1. Check self.dataloader's length. It should be 1 because all dataset item have same video_id
        2. Check self.dataloader.dataset's length. It should be 20 (self.data_len - # fo empty frame)
        3. Check self.dataloader.original_dataset's length. It should be 40 (self.data_len)
    2. Test __getitem__
    <Steps>
        1. Check len(output) == clip_len(8)
        2. Check frame_indices == [0, 0, 0, 0, 0, 2, 4, 6]. It comes from setting indices rule.
    3. Test add_prediciton
    <Steps>
        1. Check only center frame's annotations are updated
    """

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.data_len = 40
        self.labels = generate_labels(1, Domain.ACTION_DETECTION)
        self.dataset = generate_action_det_otx_dataset(1, self.data_len, self.labels)[0]
        self.dataloader = ActionOVDetDataLoader(self.dataset, 8, 256, 256)

    @e2e_pytest_unit
    def test_init(self) -> None:
        """Test __init__ function."""

        assert len(self.dataloader) == 20
        assert len(self.dataloader.original_dataset) == 40

        sample = self.dataloader.dataset[0]
        metadata = sample.get_metadata()[0].data
        assert "start_index" in metadata.metadata
        assert "timestamp_start" in metadata.metadata
        assert "timestamp_end" in metadata.metadata

    @e2e_pytest_unit
    def test_getitem(self) -> None:
        """Test __getitem__ function."""

        outs = self.dataloader[0]
        frame_indices = []
        for out in outs:
            frame_idx = out.get_metadata()[0].data.frame_idx
            frame_indices.append(frame_idx)
        assert len(outs) == 8
        assert frame_indices == [0, 0, 0, 0, 0, 2, 4, 6]

    @e2e_pytest_unit
    def test_add_prediction(self) -> None:
        """Test add_prediciton function."""

        prediction = AnnotationSceneEntity(
            annotations=[
                Annotation(
                    Rectangle.generate_full_box(), [ScoredLabel(LabelEntity("2", Domain.ACTION_DETECTION, id=ID(2)))]
                )
            ],
            kind=AnnotationSceneKind.ANNOTATION,
        )
        items = self.dataset.with_empty_annotations()._items
        self.dataloader.add_prediction(items, prediction)
        assert len(items[0].get_annotations()) == 0
        assert len(items[len(items) // 2].get_annotations()) >= 1



import pytest
import torch

from otx.algorithms.segmentation.adapters.mmseg.models.losses.detcon_loss import (
    DetConLoss,
    manual_cross_entropy,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
@pytest.mark.parametrize(
    "logits,labels,weight,expected",
    [
        (
            torch.Tensor([[[-1.2786, 2.1673, -1.0000e09, 6.2263], [-6.7169, -7.1416, 7.3364, -1.0000e09]]]),
            torch.Tensor([[[1.0, 0.0, 0.0, 0.0], [0.0, 1.0, 0.0, 0.0]]]),
            torch.Tensor([[1.0, 1.0]]),
            torch.tensor(11.0003),
        ),
        (
            torch.Tensor(
                [[[4.3260e-01, 1.6511e00, -1.0000e09, -7.9047e-01], [9.4888e-01, 2.6476e00, -3.7093e-01, -1.0000e09]]]
            ),
            torch.Tensor([[[1.0, 0.0, 0.0, 0.0], [0.0, 1.0, 0.0, 0.0]]]),
            torch.Tensor([[1.0, 1.0]]),
            torch.tensor(0.8755),
        ),
    ],
)
def test_manual_cross_entropy(logits, labels, weight, expected):
    results = manual_cross_entropy(logits, labels, weight)

    assert torch.allclose(results, expected)


@e2e_pytest_unit
@pytest.mark.parametrize(
    "inputs,expected",
    [
        (
            {
                "pred1": torch.Tensor([[[-0.0461, 0.1686, -0.1898, -0.1727], [-0.0609, -0.0948, -0.0463, -0.1461]]]),
                "pred2": torch.Tensor([[[-0.0811, -0.0115, -0.1901, -0.0227], [-0.0230, 0.0898, -0.1495, -0.0360]]]),
                "target1": torch.Tensor([[[-0.7734, 0.1523, 0.4729, -2.2324], [-0.3711, 0.2083, 0.0198, -1.9111]]]),
                "target2": torch.Tensor([[[-1.0625, 1.4160, 0.7988, 1.3291], [-0.2479, 0.9683, 0.0859, 0.4832]]]),
                "pind1": torch.Tensor([[1, 0]]),
                "pind2": torch.Tensor([[1, 0]]),
                "tind1": torch.Tensor([[1, 0]]),
                "tind2": torch.Tensor([[1, 0]]),
            },
            torch.tensor(11.8758),
        )
    ],
)
def test_detcon_loss(mocker, inputs, expected):
    mocker.patch("torch.cuda.device_count", return_value=0)
    detcon_loss = DetConLoss()

    results = detcon_loss(**inputs)

    assert "loss" in results
    assert torch.allclose(results["loss"], expected)



"""Data Pipeline of SSD model for Detection Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

__dataset_type = "CocoDataset"
__img_size = (864, 864)
__img_norm_cfg = dict(mean=[0, 0, 0], std=[255, 255, 255], to_rgb=True)

train_pipeline = [
    dict(type="LoadImageFromFile", to_float32=True),
    dict(type="LoadAnnotations", with_bbox=True),
    dict(
        type="PhotoMetricDistortion",
        brightness_delta=32,
        contrast_range=(0.5, 1.5),
        saturation_range=(0.5, 1.5),
        hue_delta=18,
    ),
    dict(type="MinIoURandomCrop", min_ious=(0.1, 0.3, 0.5, 0.7, 0.9), min_crop_size=0.1),
    dict(type="Resize", img_scale=__img_size, keep_ratio=False),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="RandomFlip", flip_ratio=0.5),
    dict(type="DefaultFormatBundle"),
    dict(type="Collect", keys=["img", "gt_bboxes", "gt_labels"]),
]
test_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(
        type="MultiScaleFlipAug",
        img_scale=__img_size,
        flip=False,
        transforms=[
            dict(type="Resize", keep_ratio=False),
            dict(type="Normalize", **__img_norm_cfg),
            dict(type="ImageToTensor", keys=["img"]),
            dict(type="Collect", keys=["img"]),
        ],
    ),
]
data = dict(
    samples_per_gpu=10,
    workers_per_gpu=4,
    train=dict(
        type=__dataset_type,
        ann_file="data/coco/annotations/instances_train2017.json",
        img_prefix="data/coco/train2017",
        pipeline=train_pipeline,
    ),
    val=dict(
        type=__dataset_type,
        ann_file="data/coco/annotations/instances_val2017.json",
        img_prefix="data/coco/val2017",
        test_mode=True,
        pipeline=test_pipeline,
    ),
    test=dict(
        type=__dataset_type,
        ann_file="data/coco/annotations/instances_val2017.json",
        img_prefix="data/coco/val2017",
        test_mode=True,
        pipeline=test_pipeline,
    ),
)



import pytest
import torch
import torch.nn as nn

from otx.algorithms.classification.adapters.mmcls import ConstrastiveHead
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestConstrastiveHead:
    @pytest.fixture(autouse=True)
    def setup(self, monkeypatch) -> None:
        class MockNeck(nn.Sequential):
            def __init__(self):
                super().__init__(nn.Linear(2, 2, bias=False))

            def init_weights(self, init_linear=None):
                for module in self.modules():
                    if hasattr(module, "weight") and module.weight is not None:
                        nn.init.constant_(module.weight, 1)

        def build_mock_neck(*args, **kwargs):
            return MockNeck()

        monkeypatch.setattr(
            "otx.algorithms.classification.adapters.mmcls.models.heads.contrastive_head.build_neck", build_mock_neck
        )

        self.inputs = torch.Tensor([[1.0, 2.0], [3.0, 4.0]])
        self.targets = torch.Tensor([[2.0, 3.0], [4.0, 5.0]])

    @e2e_pytest_unit
    def test_forward(self) -> None:
        """Test forward function."""
        contrastive_head = ConstrastiveHead(predictor={})
        contrastive_head.init_weights()

        result = contrastive_head(self.inputs, self.targets)
        expected_result = {"loss": torch.tensor(0.0255)}

        assert torch.allclose(result["loss"], expected_result["loss"], rtol=1e-4, atol=1e-4)

    @e2e_pytest_unit
    def test_forward_no_size_average(self) -> None:
        """Test forward function without size averaging."""
        contrastive_head = ConstrastiveHead(predictor={}, size_average=False)
        contrastive_head.init_weights()

        result = contrastive_head(self.inputs, self.targets)
        expected_result = {"loss": torch.tensor(0.0511)}

        assert torch.allclose(result["loss"], expected_result["loss"], rtol=1e-4, atol=1e-4)



"""Unload Interface.

This module contains the interface class for tasks to be notified when the task does not need to be loaded anymore.
"""


# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc


class IUnload(metaclass=abc.ABCMeta):
    """Interface to provide unload functionality.

    This interface can be implemented by a task, if the task wants to be notified when
    the task is not needed by the pipeline anymore.
    This allows to clear GPU and system memory resources for example.
    """

    @abc.abstractmethod
    def unload(self):
        """Unload task.

        Unload any resources which have been used by the task.

        It is acceptable to restart the server as a last resort strategy if unloading the resources is too difficult.
        """
        raise NotImplementedError



"""Tests for anomaly classification with OTX CLI"""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os

import pytest

from otx.cli.registry import Registry
from tests.test_suite.e2e_test_system import e2e_pytest_component
from tests.test_suite.run_test_command import (
    nncf_eval_openvino_testing,
    nncf_eval_testing,
    nncf_export_testing,
    nncf_optimize_testing,
    nncf_validate_fq_testing,
    otx_demo_deployment_testing,
    otx_demo_openvino_testing,
    otx_demo_testing,
    otx_deploy_openvino_testing,
    otx_eval_deployment_testing,
    otx_eval_openvino_testing,
    otx_eval_testing,
    otx_export_testing,
    otx_train_testing,
    pot_eval_testing,
    pot_optimize_testing,
    pot_validate_fq_testing,
)

args = {
    "--train-data-roots": "tests/assets/anomaly/shapes/train",
    "--val-data-roots": "tests/assets/anomaly/shapes/test",
    "--test-data-roots": "tests/assets/anomaly/shapes/test",
    "--input": "tests/assets/anomaly/shapes/test/hexagon",
    "train_params": [],
}

otx_dir = os.getcwd()

templates = Registry("otx/algorithms").filter(task_type="ANOMALY_CLASSIFICATION").templates
templates_ids = [template.model_template_id for template in templates]


class TestToolsAnomalyClassification:
    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_train(self, template, tmp_dir_path):
        otx_train_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_export(self, template, tmp_dir_path):
        otx_export_testing(template, tmp_dir_path)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval(self, template, tmp_dir_path):
        otx_eval_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval_openvino(self, template, tmp_dir_path):
        otx_eval_openvino_testing(template, tmp_dir_path, otx_dir, args, threshold=0.0)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_demo(self, template, tmp_dir_path):
        otx_demo_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_demo_openvino(self, template, tmp_dir_path):
        otx_demo_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_deploy_openvino(self, template, tmp_dir_path):
        otx_deploy_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval_deployment(self, template, tmp_dir_path):
        otx_eval_deployment_testing(template, tmp_dir_path, otx_dir, args, threshold=0.0)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_demo_deployment(self, template, tmp_dir_path):
        otx_demo_deployment_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_optimize(self, template, tmp_dir_path):
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_optimize_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_export(self, template, tmp_dir_path):
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_export_testing(template, tmp_dir_path)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_validate_fq(self, template, tmp_dir_path):
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_validate_fq_testing(template, tmp_dir_path, otx_dir, "anomaly", type(self).__name__)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_eval(self, template, tmp_dir_path):
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_eval_testing(template, tmp_dir_path, otx_dir, args, threshold=0.0001)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_eval_openvino(self, template, tmp_dir_path):
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_eval_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_pot_optimize(self, template, tmp_dir_path):
        pot_optimize_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_pot_validate_fq(self, template, tmp_dir_path):
        pot_validate_fq_testing(template, tmp_dir_path, otx_dir, "anomaly", type(self).__name__)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_pot_eval(self, template, tmp_dir_path):
        pot_eval_testing(template, tmp_dir_path, otx_dir, args)



"""Tests the methods in the train task."""

# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from copy import deepcopy

from torch import nn

from otx.algorithms.anomaly.tasks.train import TrainingTask
from otx.api.entities.train_parameters import TrainParameters
from tests.unit.algorithms.anomaly.helpers.utils import create_task_environment


class TestTrainTask:
    """Tests the methods in the train task."""

    def _compare_state_dict(self, model1: nn.Module, model2: nn.Module):
        """Compares the state dict of two models."""
        state_dict1 = model1.state_dict()
        state_dict2 = model2.state_dict()
        for (key1, param1), (key2, param2) in zip(state_dict1.items(), state_dict2.items()):
            assert key1 == key2
            if not param1.data.isnan().any() and "bn" not in key1:
                assert param1.data.allclose(param2.data)

    def test_train_and_load(self, tmpdir, setup_task_environment):
        """Tests the train method and check if it can be loaded correctly."""
        root = str(tmpdir.mkdir("anomaly_training_test"))

        # Get task environment
        setup_task_environment = deepcopy(setup_task_environment)  # since fixture is mutable
        task_environment = setup_task_environment.task_environment
        output_model = setup_task_environment.output_model
        dataset = setup_task_environment.dataset
        task_type = setup_task_environment.task_type

        train_task = TrainingTask(task_environment, output_path=root)
        train_task.train(dataset, output_model, TrainParameters())
        train_task.model = train_task.load_model(output_model)

        # create a new output_model and load from task environment
        # check if the loaded model is the same as the trained model
        new_task_environment = create_task_environment(dataset, task_type)
        new_task_environment.model = output_model
        new_task = TrainingTask(new_task_environment, output_path=root)  # should load the model from the output_model
        self._compare_state_dict(train_task.model, new_task.model)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import inspect
from collections import OrderedDict
from copy import deepcopy
from dataclasses import asdict
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

import _collections_abc
import networkx as nx
from openvino.pyopenvino import Model

from otx.mpa.utils.logger import get_logger

from ..ops.op import Operation
from ..utils import convert_op_to_torch, get_op_name

logger = get_logger()


class SortedDictKeysView(_collections_abc.KeysView):
    def __repr__(self):
        return f"{self.__class__.__name__}({[i for i in self._mapping]})"

    def __reversed__(self):
        yield from reversed(self._mapping)


class SortedDictValuesView(_collections_abc.ValuesView):
    def __repr__(self):
        return f"{self.__class__.__name__}({[self._mapping[i] for i in self._mapping]})"

    def __reversed__(self):
        for key in reversed(self._mapping):
            yield self._mapping[key]


class SortedDictItemsView(_collections_abc.ItemsView):
    def __repr__(self):
        return f"{self.__class__.__name__}({[(i, self._mapping[i]) for i in self._mapping]})"

    def __reversed__(self):
        for key in reversed(self._mapping):
            yield (key, self._mapping[key])


class NOOP:
    pass


class SortedDict(dict):
    def __init__(self, sort_key, *args, **kwargs):
        self._sort_key = sort_key
        self._sorted_keys = []
        super().__init__(self, *args, **kwargs)

    def __setitem__(self, key, value):
        assert len(value) == 1
        edge_key, edge_attr = next(iter(value.items()))
        sort_value = float("inf") if self._sort_key not in edge_attr else edge_attr[self._sort_key]
        self._sorted_keys.append([sort_value, key, edge_key])
        self._sorted_keys.sort(key=lambda x: x[0])
        if key in self:
            assert edge_key not in self[key]
            self[key].update(value)
        else:
            super().__setitem__(key, value)

    def __delitem__(self, key):
        super().__delitem__(key)
        for i, (_, key_in, _) in enumerate(self._sorted_keys):
            if key_in == key:
                break
        self._sorted_keys.pop(i)

    def __iter__(self):
        for _, key, _ in self._sorted_keys:
            yield key

    def __reversed__(self):
        for _, key, _ in self._sorted_keys[::-1]:
            yield key

    def __repr__(self):
        if not len(self):
            return "{}"
        repr = "{"
        for _, key, _ in self._sorted_keys:
            repr += f"{key}: {self[key]}, "
        repr = repr[:-2]
        repr += "}"
        return repr

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls(self._sort_key)
        memo[id(self)] = result
        for k, v in self.items():
            result[k] = deepcopy(v, memo)
        return result

    def clear(self):
        super().clear()
        self._sorted_keys = []

    def pop(self, key, default=NOOP()):
        if isinstance(default, NOOP):
            value = super().pop(key)
        else:
            value = super().pop(key, default)

        for i, (_, key_in, _) in enumerate(self._sorted_keys):
            if key_in == key:
                break
        self._sorted_keys.pop(i)

        return value

    def popitem(self):
        raise NotImplementedError

    @staticmethod
    def fromkeys(iterable, value=None):
        raise NotImplementedError

    def keys(self):
        return SortedDictKeysView(self)

    def values(self):
        return SortedDictValuesView(self)

    def items(self):
        return SortedDictItemsView(self)


class SortedDictHelper(dict):
    def __init__(self, sort_key=None, *args, **kwargs):
        self._sort_key = sort_key
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        super().__setitem__(key, SortedDict(self._sort_key))
        for v_key, v_value in value.items():
            self[key][v_key] = v_value


class Graph(nx.MultiDiGraph):
    adjlist_outer_dict_factory = SortedDictHelper

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._adj = self.adjlist_outer_dict_factory("out_port")
        self._pred = self.adjlist_outer_dict_factory("in_port")
        self._succ = self._adj

        self._normalize_nodes = []

    @staticmethod
    def from_ov(ov_model: Model) -> "Graph":
        graph = Graph()

        ov_ops = ov_model.get_ordered_ops()
        ops_dict = OrderedDict()
        parents_dict = {}
        children_dict = {}

        for ov_op in ov_ops:
            op_name = get_op_name(ov_op)
            node = convert_op_to_torch(ov_op)

            graph.add_node(
                node,
                name=op_name,
                type=node.type,
                version=node.version,
                attrs=node.attrs,
            )
            ops_dict[op_name] = node

            children_dict[op_name] = []
            for out_port in ov_op.outputs():
                out_port_id = out_port.get_index()
                for in_port in out_port.get_target_inputs():
                    in_port_id = in_port.get_index()
                    children_dict[op_name].append([out_port_id, in_port_id, get_op_name(in_port.get_node())])

            parents_dict[op_name] = []
            for in_port in ov_op.inputs():
                in_port_id = in_port.get_index()
                out_port = in_port.get_source_output()
                out_port_id = out_port.get_index()
                parents_dict[op_name].append([out_port_id, in_port_id, get_op_name(out_port.get_node())])

        # validate graph
        for node, children in children_dict.items():
            for _, _, child in children:
                assert node in [i[-1] for i in parents_dict[child]], f"{node} is not a parent of {child}"
        for node, parents in parents_dict.items():
            for _, _, parent in parents:
                assert node in [i[-1] for i in children_dict[parent]], f"{node} is not a child of {parent}"

        # add edges
        for src, tgts in children_dict.items():
            for out_port_id, in_port_id, tgt in tgts:
                graph.add_edge(
                    ops_dict[src],
                    ops_dict[tgt],
                    in_port=in_port_id,
                    out_port=out_port_id,
                )

        # freeze normalization nodes
        graph._freeze_normalize_nodes()

        return graph

    def get_edge_data(self, node_from: Operation, node_to: Operation, default=None) -> Optional[List[Dict[Any, Any]]]:
        edge_data = super().get_edge_data(node_from, node_to, None, default)
        if edge_data is not None:
            return list(edge_data.values())
        else:
            return None

    def remove_node(self, node: Operation, keep_connect: bool = False):
        edges_to_keep = []
        if keep_connect:
            predecessors = [predecessor for predecessor in self.predecessors(node) if predecessor.type != "Constant"]
            if predecessors:
                assert len(predecessors) == 1
                predecessor = predecessors[0]

                for successor in self.successors(node):
                    for predecessor_ in self.predecessors(successor):
                        edges_attrs = self.get_edge_data(predecessor_, successor)
                        assert len(edges_attrs) == 1
                        if predecessor_ == node:
                            for edge_attrs in edges_attrs:
                                edges_to_keep.append([predecessor, successor, edge_attrs])

        super().remove_node(node)
        for edge in edges_to_keep:
            node_from, node_to, attrs = edge
            self.add_edge(node_from, node_to, **attrs)

    def replace_node(self, old_node: Operation, new_node: Operation):
        edges = []
        for successor in self.successors(old_node):
            for edge_attrs in self.get_edge_data(old_node, successor):
                edges.append((new_node, successor, edge_attrs))
        for predecessor in self.predecessors(old_node):
            for edge_attrs in self.get_edge_data(predecessor, old_node):
                edges.append((predecessor, new_node, edge_attrs))

        self.remove_node(old_node)

        for edge in edges:
            node_from, node_to, attrs = edge
            self.add_edge(node_from, node_to, **attrs)

    def add_edge(
        self,
        node_from: Operation,
        node_to: Operation,
        out_port: Optional[int] = None,
        in_port: Optional[int] = None,
        **kwargs,
    ):
        if node_from not in self:
            self.add_node(node_from)

        if node_to not in self:
            self.add_node(node_to)

        if out_port is None:
            out_port = 0

        if in_port is None:
            occupied = [
                edge["in_port"]
                for predecessor in self.predecessors(node_to)
                for edge in self.get_edge_data(predecessor, node_to)
            ]
            assert len(occupied) == len(set(occupied))
            if occupied:
                for i in range(max(occupied)):
                    if i not in occupied:
                        in_port = i
                        break
            if in_port is None:
                in_port = len(occupied)

        # validate in_port
        spec = inspect.getfullargspec(node_to.forward)
        if spec.varargs is None:
            valid_range = list(range(len(spec.args[1:])))
            if in_port not in valid_range:
                raise ValueError(f"in_port {in_port} is not in valid range {valid_range} " f"for {node_to.name}.")
        occupied = []
        predecessors = list(self.predecessors(node_to))
        if predecessors:
            occupied = [
                edge["in_port"] for predecessor in predecessors for edge in self.get_edge_data(predecessor, node_to)
            ]
            assert len(occupied) == len(set(occupied))
        if occupied:
            if in_port in occupied:
                raise ValueError(f"in_port {in_port} is occupied for {node_to.name}.")

        # out_port validation is not able to do

        # add edge
        key = f"{in_port}{out_port}"
        super().add_edge(node_from, node_to, key=key, in_port=in_port, out_port=out_port, **kwargs)

    def predecessors(
        self,
        node: Operation,
        with_edge_data: bool = False,
    ) -> Generator[Union[Tuple[Operation, Optional[List]], Operation], None, None]:
        for predecessor in super().predecessors(node):
            if with_edge_data:
                yield (predecessor, self.get_edge_data(predecessor, node))
            else:
                yield predecessor

    def successors(
        self,
        node: Operation,
        with_edge_data: bool = False,
    ) -> Generator[Union[Tuple[Operation, Optional[List]], Operation], None, None]:
        for successor in super().successors(node):
            if with_edge_data:
                yield (successor, self.get_edge_data(node, successor))
            else:
                yield successor

    def get_nodes_by_types(self, types: List[str]) -> List[Operation]:
        found = []
        for node in self.topological_sort():
            if node.type in types:
                found.append(node)
        return found

    def bfs(
        self, node: Operation, reverse: bool = False, depth_limit: Optional[int] = None
    ) -> Generator[Union[Tuple[Operation, Operation], Tuple[Operation, Tuple[Operation]]], None, None]:
        if reverse:
            for s, t in nx.bfs_edges(self, node, reverse=True, depth_limit=depth_limit):
                yield (t, s)
        else:
            parent = node
            children = []
            for p, c in nx.bfs_edges(self, node, depth_limit=depth_limit):
                if p == parent:
                    children.append(c)
                    continue
                yield (parent, tuple(children))
                children = [c]
                parent = p
            if children:
                yield (parent, tuple(children))

    #  def dfs(self, node: Operation, forward=True, depth_limit=None):
    #      if forward:
    #          return nx.dfs_successors(self, node, depth_limit)
    #      else:
    #          return nx.dfs_predecessors(self, node, depth_limit)

    def get_nodes_by_type_pattern(self, pattern: List[str], start_node: Optional[Operation] = None, reverse=False):
        if len(pattern) < 1:
            raise ValueError(f"pattern must be longer than 2 but {len(pattern)} is given")
        pattern_pairs = [pattern[i : i + 2] for i in range(len(pattern) - 1)]

        if start_node is None:
            if reverse:
                start_node = list(self.topological_sort())[-1]
            else:
                start_node = list(self.topological_sort())[0]

        founds = []
        start_nodes = [start_node]
        for pattern_pair in pattern_pairs:
            found_ = {start_node: None for start_node in start_nodes}
            for start_node in start_nodes:
                for s, ts in self.bfs(start_node, reverse, 1):
                    if not isinstance(ts, tuple):
                        ts = (ts,)
                    for t in ts:
                        if [s.type, t.type] == pattern_pair:
                            if reverse:
                                found_[t] = s
                            else:
                                found_[s] = t
            if founds:
                pop_indices = []
                for i, found in enumerate(founds):
                    last_node = found[-1]
                    if last_node not in found_.keys():
                        pop_indices.append(i)
                    else:
                        founds[i].append(found_[last_node])
                for pop_idx in pop_indices[::-1]:
                    founds.pop(pop_idx)
            else:
                founds = [[s, t] for s, t in found_.items() if s is not None and t is not None]
            start_nodes = [found[-1] for found in founds]
        return founds

    def _freeze_normalize_nodes(self):  # noqa: C901
        invariant_types = ["Transpose", "Convert"]

        def test_constant(node):
            constant_nodes = [node_ for node_ in self.predecessors(node) if node_.type == "Constant"]
            if len(constant_nodes) != 1:
                return False
            constant_value = constant_nodes[0].data.squeeze()
            if constant_value.dim() > 1 or constant_value.numel() not in [1, 3]:
                return False
            return True

        def get_nodes_by_type_from_node(
            node,
            type,
            ignore_types=[],
            reverse=False,
            depth_limit=-1,
        ):
            func = self.successors
            if reverse:
                func = self.predecessors

            candidates = [(i, 1) for i in func(node)]
            found = []
            for candidate, cur_depth in candidates:
                if depth_limit > -1 and cur_depth > depth_limit:
                    break
                if candidate.type == type:
                    found.append(candidate)
                elif candidate.type in ignore_types:
                    candidates.extend([(i, cur_depth + 1) for i in func(candidate)])
            return found

        def find_multiply_add(node):
            scale_node = None
            mean_node = None

            scale_nodes = get_nodes_by_type_from_node(node, "Multiply", invariant_types)
            if scale_nodes and len(scale_nodes) == 1:
                scale_node = scale_nodes[0]
                if not test_constant(scale_node):
                    scale_node = None

            node = scale_node if scale_node is not None else node
            mean_nodes = get_nodes_by_type_from_node(node, "Add", invariant_types)
            if mean_nodes and len(mean_nodes) == 1:
                mean_node = mean_nodes[0]
                if not test_constant(mean_node):
                    mean_node = None

            return (scale_node, mean_node)

        def find_subtract_divide(node):
            mean_node = None
            scale_node = None

            mean_nodes = get_nodes_by_type_from_node(node, "Subtract", invariant_types)
            if mean_nodes and len(mean_nodes) == 1:
                mean_node = mean_nodes[0]
                if not test_constant(mean_node):
                    mean_node = None

            node = mean_node if mean_node is not None else node
            scale_nodes = get_nodes_by_type_from_node(node, "Divide", invariant_types)
            if scale_nodes and len(scale_nodes) == 1:
                scale_node = scale_nodes[0]
                if not test_constant(scale_node):
                    scale_node = None

            return (mean_node, scale_node)

        def find_subtract_multiply(node):
            mean_node = None
            scale_node = None

            mean_nodes = get_nodes_by_type_from_node(node, "Subtract", invariant_types)
            if mean_nodes and len(mean_nodes) == 1:
                mean_node = mean_nodes[0]
                if not test_constant(mean_node):
                    mean_node = None

            node = mean_node if mean_node is not None else node
            scale_nodes = get_nodes_by_type_from_node(node, "Multiply", invariant_types)
            if scale_nodes and len(scale_nodes) == 1:
                scale_node = scale_nodes[0]
                if not test_constant(scale_node):
                    scale_node = None

            return (mean_node, scale_node)

        for node in self:
            if node.type != "Parameter":
                continue

            # others
            found = find_multiply_add(node)
            if not any(found):
                # onnx, paddle
                found = find_subtract_divide(node)
                found_ = find_subtract_multiply(node)
                if len([i for i in found if i is not None]) < len([i for i in found_ if i is not None]):
                    found = found_

            if not all([i is not None for i in found]):
                continue

            self._normalize_nodes.append(found)
            for normalize_node in found:
                if normalize_node is not None:
                    constant_node = [node_ for node_ in self.predecessors(normalize_node) if node_.type == "Constant"][
                        0
                    ]
                    attrs = constant_node.attrs
                    attrs = asdict(attrs)
                    attrs["is_parameter"] = False
                    new_constant_node = constant_node.__class__(
                        constant_node.name,
                        data=constant_node.data.data,
                        **attrs,
                    )
                    self.replace_node(constant_node, new_constant_node)

    def remove_normalize_nodes(self):
        for nodes in self._normalize_nodes:
            first_node, second_node = nodes

            if first_node is None:
                first_node = second_node
            elif second_node is None:
                second_node = first_node
            self.remove_node(first_node, keep_connect=True)
            logger.info(f"Remove normalize node {first_node.name}")
            try:
                self.remove_node(second_node, keep_connect=True)
                logger.info(f"Remove normalize node {second_node.name}")
            except Exception:
                pass
        self._normalize_nodes = []

    def topological_sort(self):
        return nx.topological_sort(self)

    def has_path(self, node_from: Operation, node_to: Operation):
        return nx.has_path(self, node_from, node_to)

    def clean_up(
        self,
        nodes_to_keep: List[Operation] = [],
        remove_sub_components: bool = True,
    ):
        if remove_sub_components:
            # clean up sub components
            components = list(nx.connected_components(self.to_undirected()))
            if nodes_to_keep:
                for component in components:
                    if not set(nodes_to_keep).intersection(component):
                        super().remove_nodes_from(list(component))
            else:
                components.sort(key=len, reverse=True)
                # keep largest one only
                components.pop(0)
                for component in components:
                    super().remove_nodes_from(list(component))

        # clean up isolated node
        for node in nx.isolates(self):
            if node not in nodes_to_keep:
                super().remove_node(node)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .inferrer import IncrDetectionInferrer
from .trainer import IncrDetectionTrainer

__all__ = ["IncrDetectionTrainer", "IncrDetectionInferrer"]



"""Model configuration of YOLOX model for Detection Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

_base_ = ["../../base/models/detector.py"]

model = dict(
    type="CustomYOLOX",
    backbone=dict(type="CSPDarknet", deepen_factor=0.33, widen_factor=0.375, out_indices=(2, 3, 4)),
    neck=dict(type="YOLOXPAFPN", in_channels=[96, 192, 384], out_channels=96, num_csp_blocks=1),
    bbox_head=dict(type="CustomYOLOXHead", num_classes=80, in_channels=96, feat_channels=96),
    train_cfg=dict(assigner=dict(type="SimOTAAssigner", center_radius=2.5)),
    # In order to align the source code, the threshold of the val phase is
    # 0.01, and the threshold of the test phase is 0.001.
    test_cfg=dict(score_thr=0.01, nms=dict(type="nms", iou_threshold=0.65), max_per_img=100),
)
load_from = "https://storage.openvinotoolkit.org/repositories/openvino_training_extensions\
/models/object_detection/v2/yolox_tiny_8x8.pth"

fp16 = dict(loss_scale=512.0)
ignore = False



"""Implementation of EfficientNetV2.

Original papers:
- 'EfficientNetV2: Smaller Models and Faster Training,' https://arxiv.org/abs/2104.00298,
- 'Adversarial Examples Improve Image Recognition,' https://arxiv.org/abs/1911.09665.
"""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=unused-argument, invalid-name

import os

import timm
from mmcv.runner import load_checkpoint
from torch import nn

from otx.mpa.utils.logger import get_logger

from ..builder import BACKBONES

logger = get_logger()

PRETRAINED_ROOT = "https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-effv2-weights/"
pretrained_urls = {
    "efficientnetv2_s_21k": PRETRAINED_ROOT + "tf_efficientnetv2_s_21k-6337ad01.pth",
    "efficientnetv2_s_1k": PRETRAINED_ROOT + "tf_efficientnetv2_s_21ft1k-d7dafa41.pth",
}

NAME_DICT = {
    "mobilenetv3_large_21k": "mobilenetv3_large_100_miil_in21k",
    "mobilenetv3_large_1k": "mobilenetv3_large_100_miil",
    "tresnet": "tresnet_m",
    "efficientnetv2_s_21k": "tf_efficientnetv2_s_in21k",
    "efficientnetv2_s_1k": "tf_efficientnetv2_s_in21ft1k",
    "efficientnetv2_m_21k": "tf_efficientnetv2_m_in21k",
    "efficientnetv2_m_1k": "tf_efficientnetv2_m_in21ft1k",
    "efficientnetv2_b0": "tf_efficientnetv2_b0",
}


class TimmModelsWrapper(nn.Module):
    """Timm model wrapper."""

    def __init__(self, model_name, pretrained=False, pooling_type="avg", **kwargs):
        super().__init__(**kwargs)
        self.model_name = model_name
        self.pretrained = pretrained
        if model_name in ["mobilenetv3_large_100_miil_in21k", "mobilenetv3_large_100_miil"]:
            self.is_mobilenet = True
        else:
            self.is_mobilenet = False

        self.model = timm.create_model(NAME_DICT[self.model_name], pretrained=pretrained, num_classes=1000)
        if self.pretrained:
            logger.info(f"init weight - {pretrained_urls[self.model_name]}")
        self.model.classifier = None  # Detach classifier. Only use 'backbone' part in mpa.
        self.num_head_features = self.model.num_features
        self.num_features = self.model.conv_head.in_channels if self.is_mobilenet else self.model.num_features
        self.pooling_type = pooling_type

    def forward(self, x, **kwargs):
        """Forward."""
        y = self.extract_features(x)
        return y

    def extract_features(self, x):
        """Extract features."""
        if self.is_mobilenet:
            x = self.model.conv_stem(x)
            x = self.model.bn1(x)
            x = self.model.act1(x)
            y = self.model.blocks(x)
            return y
        return self.model.forward_features(x)

    def get_config_optim(self, lrs):
        """Get optimizer configs."""
        parameters = [
            {"params": self.model.named_parameters()},
        ]
        if isinstance(lrs, list):
            assert len(lrs) == len(parameters)
            for lr, param_dict in zip(lrs, parameters):
                param_dict["lr"] = lr
        else:
            assert isinstance(lrs, float)
            for param_dict in parameters:
                param_dict["lr"] = lrs

        return parameters


@BACKBONES.register_module()
class OTXEfficientNetV2(TimmModelsWrapper):
    """EfficientNetV2 for OTX."""

    def __init__(self, version="s_21k", **kwargs):
        self.model_name = "efficientnetv2_" + version
        super().__init__(model_name=self.model_name, **kwargs)

    def init_weights(self, pretrained=None):
        """Initialize weights."""
        if isinstance(pretrained, str) and os.path.exists(pretrained):
            load_checkpoint(self, pretrained)
            logger.info(f"init weight - {pretrained}")
        elif pretrained is not None:
            load_checkpoint(self, pretrained_urls[self.model_name])
            logger.info(f"init weight - {pretrained_urls[self.model_name]}")



"""Unit test for otx.algorithms.action.adapters.mmaction.models.backbones."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmdet.models import BACKBONES as MMDET_BACKBONES

from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
def test_register_action_backbones() -> None:
    """Test register_action_backbones function.

    Since this function is called while initialization, X3D should be in mmdet backbone registry
    """

    assert "X3D" in MMDET_BACKBONES



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import math
from dataclasses import dataclass, field
from typing import List

import torch
from torch.nn import functional as F

from .builder import OPS
from .op import Attribute, Operation


@dataclass
class PadV1Attribute(Attribute):
    pad_mode: str

    def __post_init__(self):
        super().__post_init__()
        valid_pad_mode = ["constant", "edge", "reflect", "symmetric"]
        if self.pad_mode not in valid_pad_mode:
            raise ValueError(f"Invalid pad_mode {self.pad_mode}. " f"It must be one of {valid_pad_mode}.")


@OPS.register()
class PadV1(Operation[PadV1Attribute]):
    TYPE = "Pad"
    VERSION = 1
    ATTRIBUTE_FACTORY = PadV1Attribute

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pad_mode = self.get_torch_pad_mode(self.attrs.pad_mode)

    @staticmethod
    def get_torch_pad_mode(pad_mode):
        if pad_mode == "constant":
            return "constant"
        elif pad_mode == "edge":
            return "replicate"
        elif pad_mode == "reflect":
            return "reflect"
        elif pad_mode == "symmetric":
            raise NotImplementedError
        else:
            raise NotImplementedError

    @staticmethod
    def get_torch_pad_dim(pads_begin, pads_end):
        # reverse padding
        return [val for tup in zip(pads_begin[::-1], pads_end[::-1]) for val in tup]

    def forward(self, input, pads_begin, pads_end, pad_value=0):
        pads_begin = pads_begin if isinstance(pads_begin, list) else pads_begin.detach().cpu().tolist()
        pads_end = pads_end if isinstance(pads_end, list) else pads_end.detach().cpu().tolist()
        pad = self.get_torch_pad_dim(pads_begin, pads_end)
        pad = list(map(math.ceil, pad))
        return F.pad(input=input, pad=pad, mode=self._pad_mode, value=pad_value)


@dataclass
class ConcatV0Attribute(Attribute):
    axis: int


@OPS.register()
class ConcatV0(Operation[ConcatV0Attribute]):
    TYPE = "Concat"
    VERSION = 0
    ATTRIBUTE_FACTORY = ConcatV0Attribute

    def forward(self, *inputs):
        return torch.cat(inputs, self.attrs.axis)


@dataclass
class TransposeV1Attribute(Attribute):
    pass


@OPS.register()
class TransposeV1(Operation[TransposeV1Attribute]):
    TYPE = "Transpose"
    VERSION = 1
    ATTRIBUTE_FACTORY = TransposeV1Attribute

    def forward(self, input, order):
        if order.numel() == 0:
            order = list(range(input.dim()))[::-1]
        elif isinstance(order, torch.Tensor):
            order = order.detach().cpu().tolist()
        return input.permute(order)


@dataclass
class GatherV0Attribute(Attribute):
    batch_dims: int = field(default=0)


@OPS.register()
class GatherV0(Operation[GatherV0Attribute]):
    TYPE = "Gather"
    VERSION = 0
    ATTRIBUTE_FACTORY = GatherV0Attribute

    def forward(self, input, indices, axis):
        assert axis.numel() == 1
        axis = axis.squeeze()
        squeeze_axis = indices.dim() == 0

        batch_dims = self.attrs.batch_dims
        if batch_dims < 0:
            batch_dims = indices.dim() + batch_dims

        indices_shape = torch.tensor(indices.shape)
        if batch_dims < axis:
            indices = indices.reshape(*indices_shape[:batch_dims], -1)
            indices_shape = indices_shape[batch_dims:]

        if indices.dim() != input.dim():
            if indices.dim() != 0:
                while indices.dim() - 1 < axis:
                    indices = indices.unsqueeze(batch_dims)
            while indices.dim() < input.dim():
                indices = indices.unsqueeze(-1)

            repeat = []
            for i, (j, k) in enumerate(zip(input.shape, indices.shape)):
                if i == axis:
                    repeat.append(1)
                else:
                    assert j % k == 0
                    repeat.append(j // k)
            indices = indices.repeat(repeat)
        output = torch.gather(input=input, dim=axis, index=indices.type(torch.int64))

        if squeeze_axis:
            output = output.squeeze(axis)

        return output


@dataclass
class GatherV1Attribute(Attribute):
    pass


@OPS.register()
class GatherV1(Operation[GatherV1Attribute]):
    TYPE = "Gather"
    VERSION = 1
    ATTRIBUTE_FACTORY = GatherV1Attribute

    def forward(self, input, indices, axis):
        return torch.gather(input=input, dim=axis, index=indices)


@dataclass
class StridedSliceV1Attribute(Attribute):
    begin_mask: List[int]
    end_mask: List[int]
    new_axis_mask: List[int] = field(default_factory=lambda: [0])
    shrink_axis_mask: List[int] = field(default_factory=lambda: [0])
    ellipsis_mask: List[int] = field(default_factory=lambda: [0])


@OPS.register()
class StridedSliceV1(Operation[StridedSliceV1Attribute]):
    TYPE = "StridedSlice"
    VERSION = 1
    ATTRIBUTE_FACTORY = StridedSliceV1Attribute

    def forward(self, input, begin, end, stride=None):
        if sum(self.attrs.ellipsis_mask) > 0:
            raise NotImplementedError

        for i, mask in enumerate(self.attrs.begin_mask):
            if mask == 1:
                begin[i] = 0
        for i, mask in enumerate(self.attrs.end_mask):
            if mask == 1:
                end[i] = input.size(i)

        if stride is None:
            stride = torch.tensor([1 for _ in begin], dtype=begin.dtype)

        output = input
        for i, (b, e, s) in enumerate(zip(begin, end, stride)):
            length = input.size(i)

            # begin index is inclusive
            b = torch.clamp(b, -length, length - 1)
            # end index is exclusive
            e = torch.clamp(e, -length - 1, length)

            if s > 0:
                b = b + length if b < 0 else b
                e = e + length if e < 0 else e
                indices = torch.arange(b, e, s, device=input.device)
            else:
                b = b - length if b >= 0 else b
                e = e - length if e >= 0 else e
                indices = torch.arange(b, e, s, device=input.device)
                indices += length

            output = torch.index_select(output, i, indices)

        for i, mask in enumerate(self.attrs.new_axis_mask[::-1]):
            if mask == 1:
                i = abs(i - len(self.attrs.new_axis_mask) + 1)
                output = output.unsqueeze(i)

        for i, mask in enumerate(self.attrs.shrink_axis_mask[::-1]):
            if mask == 1:
                i = abs(i - len(self.attrs.new_axis_mask) + 1)
                if output.size(i) != 1:
                    raise NotImplementedError
                output = output.squeeze(i)

        return output


@dataclass
class SplitV1Attribute(Attribute):
    num_splits: int


@OPS.register()
class SplitV1(Operation[SplitV1Attribute]):
    TYPE = "Split"
    VERSION = 1
    ATTRIBUTE_FACTORY = SplitV1Attribute

    def forward(self, input, axis):
        split_size = input.shape[axis] // self.attrs.num_splits
        return torch.split(tensor=input, split_size_or_sections=split_size, dim=axis)


@dataclass
class VariadicSplitV1Attribute(Attribute):
    pass


@OPS.register()
class VariadicSplitV1(Operation[VariadicSplitV1Attribute]):
    TYPE = "VariadicSplit"
    VERSION = 1
    ATTRIBUTE_FACTORY = VariadicSplitV1Attribute

    def forward(self, input, axis, split_lengths):
        idx = [i for i, j in enumerate(split_lengths) if j == -1]
        if idx:
            assert len(idx) == 1
            idx = idx[0]
            split_lengths[idx] = input.size(axis) - sum(split_lengths) - 1
        assert input.size(axis) == sum(split_lengths)
        outputs = []
        start_idx = 0
        for length in split_lengths:
            outputs.append(
                torch.index_select(
                    input,
                    axis,
                    torch.arange(start_idx, start_idx + length, device=input.device),
                )
            )
            start_idx += length
        return tuple(outputs)


@dataclass
class ShuffleChannelsV0Attribute(Attribute):
    axis: int = field(default=1)
    group: int = field(default=1)


@OPS.register()
class ShuffleChannelsV0(Operation[ShuffleChannelsV0Attribute]):
    TYPE = "ShuffleChannels"
    VERSION = 0
    ATTRIBUTE_FACTORY = ShuffleChannelsV0Attribute

    def forward(self, input):
        #  n, c, h, w = input.shape
        assert input.dim() == 4
        origin_shape = input.shape
        origin_dim = input.dim()
        assert origin_shape[self.attrs.axis] % self.attrs.group == 0

        axis = self.attrs.axis
        axis = axis if axis >= 0 else axis + input.dim()

        target_shape = [
            0,
            self.attrs.group,
            int(origin_shape[axis] / self.attrs.group),
            0,
        ]
        if axis == 0:
            target_shape[0] = 1
            target_shape[-1] = math.prod([origin_shape[i] for i in range(axis + 1, origin_dim)])
        elif axis == input.dim() - 1:
            target_shape[0] = math.prod([origin_shape[i] for i in range(0, axis)])
            target_shape[-1] = 1
        else:
            target_shape[0] = math.prod([origin_shape[i] for i in range(0, axis)])
            target_shape[-1] = math.prod([origin_shape[i] for i in range(axis + 1, origin_dim)])

        output = input.reshape(target_shape)
        output = output.permute([0, 2, 1, 3])
        output = output.reshape(origin_shape)
        return output


@dataclass
class BroadcastV3Attribute(Attribute):
    mode: str = field(default="numpy")

    def __post_init__(self):
        super().__post_init__()
        valid_mode = ["numpy", "explicit", "bidirectional"]
        if self.mode not in valid_mode:
            raise ValueError(f"Invalid mode {self.mode}. " f"It must be one of {valid_mode}.")


@OPS.register()
class BroadcastV3(Operation[BroadcastV3Attribute]):
    TYPE = "Broadcast"
    VERSION = 3
    ATTRIBUTE_FACTORY = BroadcastV3Attribute

    def forward(self, input, target_shape, axes_mapping=None):
        if self.attrs.mode == "numpy":
            return input.expand(*target_shape)
        if self.attrs.mode == "bidirectional":
            return torch.ones(*target_shape, device=input.device) * input
        else:
            assert axes_mapping is not None
            prev = -1
            for axes in axes_mapping:
                prev += 1
                while axes - prev > 0:
                    input = input.unsqueeze(axes - 1)
                    prev += 1
            while input.dim() < len(target_shape):
                input = input.unsqueeze(-1)
            return input.expand(*target_shape)


@dataclass
class ScatterNDUpdateV3Attribute(Attribute):
    pass


@OPS.register()
class ScatterNDUpdateV3(Operation[ScatterNDUpdateV3Attribute]):
    TYPE = "ScatterNDUpdate"
    VERSION = 3
    ATTRIBUTE_FACTORY = ScatterNDUpdateV3Attribute

    def forward(self, input, indicies, updates):
        # TODO: need to verify
        if updates.numel() == 1:
            raise NotImplementedError

        # FIXME: hard-coded
        last_dim = indicies.shape[-1]
        assert last_dim == 2
        assert indicies[..., -2].sum() == 0
        input.shape[indicies.shape[-1] :]
        index = indicies[..., -1]
        for i in input.shape[indicies.shape[-1] :]:
            index = index.unsqueeze(-1).tile((i,))
        output = torch.scatter(input, 1, index, updates)

        return output


@dataclass
class ScatterUpdateV3Attribute(Attribute):
    pass


@OPS.register()
class ScatterUpdateV3(Operation[ScatterUpdateV3Attribute]):
    TYPE = "ScatterUpdate"
    VERSION = 3
    ATTRIBUTE_FACTORY = ScatterUpdateV3Attribute

    def forward(self, input, indicies, updates, axis):
        # TODO: need to verify
        axis = axis.item()

        if input.dtype != updates.dtype:
            updates = updates.type(input.dtype)

        if indicies.dim() == 0:
            assert axis == 0
            output = input
            output[indicies] = updates

        output = torch.scatter(input, axis, indicies, updates)

        return output


@dataclass
class TileV0Attribute(Attribute):
    pass


@OPS.register()
class TileV0(Operation[TileV0Attribute]):
    TYPE = "Tile"
    VERSION = 0
    ATTRIBUTE_FACTORY = TileV0Attribute

    def forward(self, input, repeats):
        return torch.tile(input, repeats.tolist())



"""Multi GPU training utility."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import logging
import os
import signal
import socket
import sys
import threading
import time
from contextlib import closing
from typing import Callable, List, Optional

import psutil
import torch
import torch.distributed as dist
import torch.multiprocessing as mp

from otx.api.configuration import ConfigurableParameters

logger = logging.getLogger(__name__)


def _get_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.bind(("", 0))
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return sock.getsockname()[1]


def get_gpu_ids(gpus: str) -> List[int]:
    """Get proper GPU indices form `--gpu` arguments.

    Given `--gpus` argument, exclude inappropriate indices and transform to list of int format.

    Args:
        gpus (str): GPU indices to use. Format should be Comma-separated indices.

    Returns:
        List[int]:
            list including proper GPU indices.
    """
    num_available_gpu = torch.cuda.device_count()
    gpu_ids = []
    for gpu_id in gpus.split(","):
        if not gpu_id.isnumeric():
            raise ValueError("--gpus argument should be numbers separated by ','.")
        gpu_ids.append(int(gpu_id))

    wrong_gpus = []
    for gpu_idx in gpu_ids:
        if gpu_idx >= num_available_gpu:
            wrong_gpus.append(gpu_idx)

    for wrong_gpu in wrong_gpus:
        gpu_ids.remove(wrong_gpu)

    if wrong_gpus:
        logger.warning(f"Wrong gpu indices are excluded. {','.join([str(val) for val in gpu_ids])} GPU will be used.")

    return gpu_ids


def set_arguments_to_argv(key: str, value: Optional[str] = None, after_params: bool = False):
    """Add arguments at proper position in `sys.argv`.

    Args:
        key (str): arguement key.
        value (str or None): argument value.
        after_params (bool): whether argument should be after `param` or not.
    """
    if key in sys.argv:
        if value is not None:
            sys.argv[sys.argv.index(key) + 1] = value
    else:
        if not after_params and "params" in sys.argv:
            sys.argv.insert(sys.argv.index("params"), key)
            if value is not None:
                sys.argv.insert(sys.argv.index("params"), value)
        else:
            if after_params and "params" not in sys.argv:
                sys.argv.append("params")
            if value is not None:
                sys.argv.extend([key, value])
            else:
                sys.argv.append(key)


class MultiGPUManager:
    """Class to manage multi GPU training.

    Args:
        train_func (Callable): model training function.
        gpu_ids (str): GPU indices to use. Format should be Comma-separated indices.
        rdzv_endpoint (str): Rendezvous endpoint for multi-node training.
        base_rank (int): Base rank of the worker.
        world_size (int): Total number of workers in a worker group.
    """

    def __init__(
        self,
        train_func: Callable,
        gpu_ids: str,
        rdzv_endpoint: str = "localhost:0",
        base_rank: int = 0,
        world_size: int = 0,
    ):
        if ":" not in rdzv_endpoint:
            raise ValueError("rdzv_endpoint must be in form <host>:<port>.")
        host, port = rdzv_endpoint.split(":")
        if port == "0":
            assert host in ["localhost", "127.0.0.1"]
            port = _get_free_port()
        rdzv_endpoint = f"{host}:{port}"

        self._train_func = train_func
        self._gpu_ids = get_gpu_ids(gpu_ids)
        self._rdzv_endpoint = rdzv_endpoint
        self._base_rank = base_rank
        if world_size == 0:
            world_size = len(self._gpu_ids)
        self._world_size = world_size
        self._main_pid = os.getpid()
        self._processes: Optional[List[mp.Process]] = None

    def is_available(self) -> bool:
        """Check multi GPU training is available.

        Returns:
            bool:
                whether multi GPU training is available.
        """
        return len(self._gpu_ids) > 1

    def setup_multi_gpu_train(
        self,
        output_path: str,
        optimized_hyper_parameters: Optional[ConfigurableParameters] = None,
    ):
        """Carry out what should be done to run multi GPU training.

        Args:
            output_path (str): output path where task output are saved.
            optimized_hyper_parameters (ConfigurableParameters or None): hyper parameters reflecting HPO result.
        """
        if optimized_hyper_parameters is not None:  # if HPO is executed, optimized HPs are applied to child processes
            self._set_optimized_hp_for_child_process(optimized_hyper_parameters)

        self._processes = self._spawn_multi_gpu_processes(output_path)

        signal.signal(signal.SIGINT, self._terminate_signal_handler)
        signal.signal(signal.SIGTERM, self._terminate_signal_handler)

        self.initialize_multigpu_train(self._rdzv_endpoint, self._base_rank, 0, self._gpu_ids, self._world_size)

        threading.Thread(target=self._check_child_processes_alive, daemon=True).start()

    def finalize(self):
        """Join all child processes."""
        if self._processes is not None:
            for p in self._processes:
                p.join()

    @staticmethod
    def initialize_multigpu_train(
        rdzv_endpoint: str,
        rank: int,
        local_rank: int,
        gpu_ids: List[int],
        world_size: int,
    ):
        """Initilization for multi GPU training.

        Args:
            rdzv_endpoint (str): Rendezvous endpoint for multi-node training.
            rank (int): The rank of worker within a worker group.
            local_rank (int): The rank of worker within a local worker group.
            gpu_ids (List[int]): list including which GPU indeces will be used.
            world_size (int): Total number of workers in a worker group.
        """

        host, port = rdzv_endpoint.split(":")
        os.environ["MASTER_ADDR"] = host
        os.environ["MASTER_PORT"] = port
        os.environ["LOCAL_WORLD_SIZE"] = str(len(gpu_ids))
        os.environ["WORLD_SIZE"] = str(world_size)
        os.environ["LOCAL_RANK"] = str(local_rank)
        os.environ["RANK"] = str(rank)
        torch.cuda.set_device(gpu_ids[local_rank])
        dist.init_process_group(backend="nccl", world_size=world_size, rank=rank)
        logger.info(f"dist info world_size = {dist.get_world_size()}, rank = {dist.get_rank()}")

    @staticmethod
    def run_child_process(
        train_func: Callable,
        output_path: str,
        rdzv_endpoint: str,
        rank: int,
        local_rank: int,
        gpu_ids: List[int],
        world_size: int,
    ):
        """Function for multi GPU child process to execute.

        Args:
            train_func (Callable): model training function.
            output_path (str): output path where task output are saved.
            rdzv_endpoint (str): Rendezvous endpoint for multi-node training.
            rank (int): The rank of worker within a worker group.
            local_rank (int): The rank of worker within a local worker group.
            gpu_ids (List[int]): list including which GPU indeces will be used.
            world_size (int): Total number of workers in a worker group.
        """

        # initialize start method
        mp.set_start_method(method=None, force=True)

        gpus_arg_idx = sys.argv.index("--gpus")
        for _ in range(2):
            sys.argv.pop(gpus_arg_idx)
        if "--enable-hpo" in sys.argv:
            sys.argv.remove("--enable-hpo")
        set_arguments_to_argv("--work-dir", output_path)
        set_arguments_to_argv("--rdzv-endpoint", rdzv_endpoint)

        MultiGPUManager.initialize_multigpu_train(rdzv_endpoint, rank, local_rank, gpu_ids, world_size)

        threading.Thread(target=MultiGPUManager.check_parent_processes_alive, daemon=True).start()

        train_func()

    @staticmethod
    def check_parent_processes_alive():
        """Check parent process is alive and if not, exit by itself."""
        cur_process = psutil.Process()
        parent = cur_process.parent()
        while True:
            time.sleep(1)
            if not parent.is_running():
                break

        logger.warning("Parent process is terminated abnormally. Process exits.")
        cur_process.kill()

    def _spawn_multi_gpu_processes(self, output_path: str) -> List[mp.Process]:
        processes = []
        ctx = mp.get_context("spawn")
        for rank in range(1, len(self._gpu_ids)):
            task_p = ctx.Process(
                target=MultiGPUManager.run_child_process,
                args=(
                    self._train_func,
                    output_path,
                    self._rdzv_endpoint,
                    self._base_rank + rank,
                    rank,
                    self._gpu_ids,
                    self._world_size,
                ),
            )
            task_p.start()
            processes.append(task_p)

        return processes

    def _terminate_signal_handler(self, signum, _frame):
        # This code prevents child processses from being killed unintentionally by proccesses forked from main process
        if self._main_pid != os.getpid():
            sys.exit()

        self._kill_child_process()

        singal_name = {2: "SIGINT", 15: "SIGTERM"}
        logger.warning(f"{singal_name[signum]} is sent. process exited.")

        sys.exit(1)

    def _kill_child_process(self):
        if self._processes is None:
            return

        for process in self._processes:
            if process.is_alive():
                logger.warning(f"Kill child process {process.pid}")
                process.kill()

    def _set_optimized_hp_for_child_process(self, hyper_parameters: ConfigurableParameters):
        set_arguments_to_argv(
            "--learning_parameters.learning_rate",
            str(hyper_parameters.learning_parameters.learning_rate),  # type: ignore[attr-defined]
            True,
        )
        set_arguments_to_argv(
            "--learning_parameters.batch_size",
            str(hyper_parameters.learning_parameters.batch_size),  # type: ignore[attr-defined]
            True,
        )

    def _check_child_processes_alive(self):
        child_is_running = True
        while child_is_running:
            time.sleep(1)
            for p in self._processes:
                if not p.is_alive() and p.exitcode != 0:
                    child_is_running = False
                    break

        logger.warning("Some of child processes are terminated abnormally. process exits.")
        self._kill_child_process()
        os.kill(self._main_pid, signal.SIGKILL)



"""Model configuration of ATSS model for Detection Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

_base_ = [
    "../../../../common/adapters/mmcv/configs/backbones/mobilenet_v2_w1.yaml",
    "../../base/models/detector.py",
]

model = dict(
    type="CustomATSS",
    neck=dict(
        type="FPN",
        in_channels=[24, 32, 96, 320],
        out_channels=64,
        start_level=1,
        add_extra_convs="on_output",
        num_outs=5,
        relu_before_extra_convs=True,
    ),
    bbox_head=dict(
        type="CustomATSSHead",
        num_classes=2,
        in_channels=64,
        stacked_convs=4,
        feat_channels=64,
        anchor_generator=dict(
            type="AnchorGenerator",
            ratios=[1.0],
            octave_base_scale=8,
            scales_per_octave=1,
            strides=[8, 16, 32, 64, 128],
        ),
        bbox_coder=dict(
            type="DeltaXYWHBBoxCoder",
            target_means=[0.0, 0.0, 0.0, 0.0],
            target_stds=[0.1, 0.1, 0.2, 0.2],
        ),
        loss_cls=dict(type="FocalLoss", use_sigmoid=True, gamma=2.0, alpha=0.25, loss_weight=1.0),
        loss_bbox=dict(type="GIoULoss", loss_weight=2.0),
        loss_centerness=dict(type="CrossEntropyLoss", use_sigmoid=True, loss_weight=1.0),
        use_qfl=False,
        qfl_cfg=dict(
            type="QualityFocalLoss",
            use_sigmoid=True,
            beta=2.0,
            loss_weight=1.0,
        ),
    ),
    train_cfg=dict(
        assigner=dict(type="ATSSAssigner", topk=9),
        allowed_border=-1,
        pos_weight=-1,
        debug=False,
    ),
    test_cfg=dict(
        nms_pre=1000,
        min_bbox_size=0,
        score_thr=0.05,
        nms=dict(type="nms", iou_threshold=0.6),
        max_per_img=100,
    ),
    backbone=dict(
        out_indices=(
            2,
            3,
            4,
            5,
        )
    ),
)

load_from = "https://storage.openvinotoolkit.org/repositories/openvino_training_extensions\
/models/object_detection/v2/mobilenet_v2-atss.pth"

fp16 = dict(loss_scale=512.0)
